/*
                       Adobe ISO PostScript Generator

	Units legend:

	Dots and Pixels are the same thing, 1 dot == 1 pixel.

	All relativity is from the upper left to the upper left.

	Paper sizes are assumed to be when dpi == 100 unless otherwise
	indicated.
 */

#ifndef LIBPS_H
#define LIBPS_H

#include <stdio.h>
#include <sys/types.h>


/*
 *	Default dots per inch:
 */
#define PSDefaultDPIX		72
#define PSDefaultDPIY		72

/*
 *	Error codes:
 */
#define PSSuccess		0
#define PSError			1
#define PSBadValue		2
#define PSErrorSystem		3
#define PSAbort			4	/* Possible user aborted. */
#define PSErrorInternal		5

/*
 *	Data types:
 */
#define PSFlags			u_int32_t


/*
 *	PostScript format parameters structure:
 */
typedef struct {

	/* These flags define which members are defined. */
#define PSFlagDPIX		(1 << 0)
#define PSFlagDPIY		(1 << 1)
#define PSFlagPaperWidth	(1 << 2)
#define PSFlagPaperHeight	(1 << 3)
#define PSFlagOutputX		(1 << 4)
#define PSFlagOutputY		(1 << 5)
	PSFlags flags;

	int	dpi_x,		/* Dots per inch. */
		dpi_y;

	int	paper_width,	/* Paper size at dpi = 100. */
		paper_height;

	int	output_x,	/* Output offset from upper left, in pixels. */
		output_y;

} PSFormatStruct;


/*	Writes PS format data to the stream fp.
 *
 *	The image data buf is given in RGBA format (4 bytes per pixel).
 *	The bytes per line bpl can be 0 which means width * 4.
 *
 *      The progress_cb specifies the progress callback, this function is
 *      called during writing and the following information is passed to
 *	it (in order); client_data, current_position, and
 *	maximum_position. Note that the position values are not relative 
 *	to any particular units.
 *
 *      A return value of 0 means keep processing while a non-zero means
 *      abort the operation.
 *
 *      Returns PSSuccess on success or PSError on error. Can also return
 *	PSBadValue if the given values in fmt are unacceptable.
 */
extern int PSWriteFilePS3RGBA(
        FILE *fp, PSFormatStruct *fmt,
        int width, int height,          /* Size of image in pixels. */
        int bpl,                        /* Bytes per pixel (or row stride). */
        const void *buf,
        void *client_data,              /* Client data for progress callback. */
        int (*progress_cb)(void *, long, long)	/* Progress callback, can be NULL. */
);

/*      Writes PS format data to the stream fp.
 *
 *      The image data buf is given in RGB format (3 bytes per pixel).
 *      The bytes per line bpl can be 0 which means width * 3.
 *
 *      The progress_cb specifies the progress callback, this function is
 *      called during writing and the following information is passed to
 *      it (in order); client_data, current_position, and
 *      maximum_position. Note that the position values are not relative
 *      to any particular units.
 *
 *      Returns PSSuccess on success or PSError on error. Can also return
 *      PSBadValue if the given values in fmt are unacceptable.
 */
extern int PSWriteFilePS3RGB(
	FILE *fp, PSFormatStruct *fmt,
	int width, int height,		/* Size of image in pixels. */
	int bpl,			/* Bytes per pixel (or row stride). */
	const void *buf,
        void *client_data,              /* Client data for progress callback. */
        int (*progress_cb)(void *, long, long)	/* Progress callback, can be NULL. */
);

/*      Writes PS format data to the stream fp.
 *
 *      The image data buf is given in greyscale format (1 byte per pixel).
 *      The bytes per line bpl can be 0 which means just width.
 *
 *      The progress_cb specifies the progress callback, this function is
 *      called during writing and the following information is passed to
 *      it (in order); client_data, current_position, and
 *      maximum_position. Note that the position values are not relative
 *      to any particular units.
 *
 *	Returns PSSuccess on success or PSError on error. Can also return
 *      PSBadValue if the given values in fmt are unacceptable.
 */
extern int PSWriteFilePS3Grey(
        FILE *fp, PSFormatStruct *fmt,
        int width, int height,          /* Size of image in pixels. */
        int bpl,                        /* Bytes per pixel (or row stride). */
        const void *buf,
        void *client_data,              /* Client data for progress callback. */
        int (*progress_cb)(void *, long, long)	/* Progress callback, can be NULL. */
);


#endif	/* LIBPS_H */
