#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <limits.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "../include/string.h"

#include "guiutils.h"
#include "fb.h"
#include "imgview.h"
#include "printopts.h"
#include "config.h"

#include "images/icon_print_48x48.xpm"
#include "images/icon_print_20x20.xpm"
#include "images/icon_cancel_20x20.xpm"

#include "images/icon_minus_16x16.xpm"
#include "images/icon_plus_16x16.xpm"
#include "images/icon_brightness_16x16.xpm"
#include "images/icon_contrast_16x16.xpm"


/* Callbacks. */
static void PrintWinPaperSizeRowDataDeleteCB(gpointer data);
static gint PrintWinCloseCB(GtkWidget *widget, GdkEvent *event, gpointer data);
static void PrintWinDestroyCB(GtkObject *object, gpointer data);
static void PrintWinSelectRowCB(
	GtkWidget *widget, gint row, gint column,
	GdkEventButton *event, gpointer data
);
static void PrintWinPrintCB(GtkWidget *widget, gpointer data);
static void PrintWinCancelCB(GtkWidget *widget, gpointer data);
static void PrintWinToggledCB(GtkWidget *widget, gpointer data);
static void PrintWinAdjustmentValueChangedCB(
	GtkAdjustment *adj, gpointer data
);
static void PrintWinDPIChangedCB(GtkWidget *widget, gpointer data);
static void PrintWinOutputEntryChangedCB(GtkWidget *widget, gpointer data);
static void PrintWinPrintFileBrowseCB(gpointer widget, gpointer data);
static gint PrintWinPreviewEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static gint PrintWinPreviewExposeCB(
	GtkWidget *widget, GdkEventExpose *expose, gpointer data
);


/* Utilities. */
static void PrintWinSyncOutputWidgets(print_win_struct *pw);
static gint PrintWinVisual(print_win_struct *pw);
static gint PrintWinOrientation(print_win_struct *pw);
static gbool PrintWinGetDPI(
	print_win_struct *pw,
        gint *dpi_x, gint *dpi_y
);
static gbool PrintWinPaperSize(
	print_win_struct *pw,
	gint *width, gint *height,
	gbool apply_dpi
);
static gbool PrintWinOutputOffsetW(
        print_win_struct *pw, gint *x, gint *y
);

static void PrintWinLandscapeImage(
	gint bpp,
	const guint8 *src_data,
	gint swidth, gint sheight, gint sbpl,
	guint8 *tar_data,
	gint tbpl
);
static void PrintWinRevertImages(
	imgview_image_struct *src_img, imgview_image_struct *tar_img
);
static void PrintWinRevertPreview(print_win_struct *pw);
static void PrintWinUpdatePreviewVisualV(
	imgview_image_struct *img, print_values_struct *v
);
static void PrintWinUpdatePreviewVisual(print_win_struct *pw);
static void PrintWinRecreatePreview(print_win_struct *pw);
static void PrintWinResizePreview(print_win_struct *pw);
static void PrintWinPreviewDraw(print_win_struct *pw);

print_win_struct *PrintWinNew(imgview_struct *iv);
void PrintWinMap(print_win_struct *pw);
void PrintWinUnmap(print_win_struct *pw);
void PrintWinDelete(print_win_struct *pw);

imgview_image_struct *PrintWinCreateImageFromValues(
        imgview_image_struct *src_img, print_values_struct *v
);
gbool PrintWinGetResponse(
        imgview_struct *iv,
	GtkWidget *ref_toplevel,        /* Reference toplevel. */
	print_values_struct *v
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)             (((x) < 0) ? ((x) * -1) : (x))

#ifndef GDK_BUTTON1
# define GDK_BUTTON1    1
#endif
#ifndef GDK_BUTTON2
# define GDK_BUTTON2    2
#endif
#ifndef GDK_BUTTON3
# define GDK_BUTTON3    3
#endif


#define PRINT_DEF_TITLE		"Print"
#define PRINT_DEF_WIDTH		640
#define PRINT_DEF_HEIGHT	480

/* Nominal size of preview drawing area, the nominal size must be
 * square in shape.
 */
#define PRINT_PREVIEW_WIDTH	200
#define PRINT_PREVIEW_HEIGHT	200

#define PRINT_CLIST_ROW_SPACING	20


/* Tooltips. */
#define PRINT_TTM_PRINT_COMMAND		"\
Command to run the print client plus any arguments"

#define PRINT_TTM_BRIGHTNESS		"Brightness"
#define PRINT_TTM_CONTRAST		"Contrast"

#define PRINT_TTM_DPI_X			"\
Dots per inch (DPI) resolution of paper media width,\
 value must match with value set on printer driver"

#define PRINT_TTM_DPI_Y			"\
Dots per inch (DPI) resolution of paper media height,\
 value must match with value set on printer driver"

#define PRINT_TTM_PREVIEW		"\
Press and hold Button1 to adjust the position, press and hold SHIFT +\
 Button1 (or Button2) to adjust the size"


/* Local GTK+ main level. */
static gint main_block_level;
/* Got positive user response if TRUE, otherwise FALSE for cancel. */
static gbool got_response;


/*
 *	Deallocates the given row data.
 */
static void PrintWinPaperSizeRowDataDeleteCB(gpointer data)
{
	gchar **strv = (gchar **)data;

	if(strv != NULL)
	{
	    /* Begin deallocating row data, the first two pointers
	     * point to allocated strings so we only need to deallocate
	     * those two.
	     */
	    g_free(strv[0]);
            g_free(strv[1]);

	    /* Deallocate row data itself. */
	    g_free(strv);
	    strv = NULL;
	}
}

/*
 *	Close "delete_event" callback.
 */
static gint PrintWinCloseCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	PrintWinCancelCB(widget, data);
	return(TRUE);
}

/*
 *	Destroy callback.
 */
static void PrintWinDestroyCB(GtkObject *object, gpointer data)
{
	return;
}

/*
 *	Paper size clist "select_row" callback.
 */
static void PrintWinSelectRowCB(
        GtkWidget *widget, gint row, gint column,
        GdkEventButton *event, gpointer data
)
{
	static gbool reenterent = FALSE;
	GtkCList *clist;
        print_win_struct *pw = (print_win_struct *)data;
        if((widget == NULL) || (pw == NULL))
            return;

	if(reenterent)
	    return;
	else
	    reenterent = TRUE;

	clist = GTK_CLIST(widget);

	if(pw->map_state)
	{
	    /* Scroll to row if not visable. */
	    if(gtk_clist_row_is_visible(clist, row) != GTK_VISIBILITY_FULL)
	        gtk_clist_moveto(
		    clist,
		    row, 0,	/* Row, column. */
		    0.5, 0.0	/* Row, column. */
	        );
	}

	/* Is this is the paper size clist? */
	if((gpointer)clist == (gpointer)pw->paper_size_clist)
	{
	    /* Since paper size has changed, we need to update the
	     * preview widget size and recreate the preview images.
	     */
	    PrintWinResizePreview(pw);
	}

	reenterent = FALSE;
}


/*
 *	Print callback.
 */
static void PrintWinPrintCB(GtkWidget *widget, gpointer data)
{
        print_win_struct *pw = (print_win_struct *)data;
        if(pw == NULL)
            return;

	/* Update response code. */
	got_response = TRUE;

	/* Break out of highest GTK+ main loop. */
	if(main_block_level > 0)
	{
	    main_block_level--;
	    gtk_main_quit();
	}
}

/*
 *	Cancel callback.
 */
static void PrintWinCancelCB(GtkWidget *widget, gpointer data)
{
        print_win_struct *pw = (print_win_struct *)data;
        if(pw == NULL)
            return;

        /* Update response code. */
        got_response = FALSE;

        /* Break out of highest GTK+ main loop. */
        if(main_block_level > 0)
        {
            main_block_level--;
            gtk_main_quit();
        }
}


/*
 *	All purpose "toggled" signal callback for all radio and
 *	toggle buttons.
 */
static void PrintWinToggledCB(GtkWidget *widget, gpointer data)
{
	static gbool reenterent = FALSE;
	GtkWidget *w;
        print_win_struct *pw = (print_win_struct *)data;
        if((widget == NULL) || (pw == NULL))
            return;

	/* We're only interested in toggled to TRUE signals. */
	if(GTK_IS_TOGGLE_BUTTON(widget) ?
	    !GTK_TOGGLE_BUTTON(widget)->active : 1
	)
	    return;


	if(reenterent)
	    return;
	else
	    reenterent = TRUE;

	/* Output to printer? */
	if(widget == pw->output_printer_radio)
	{
            w = pw->output_printer_entry;
            if(w != NULL)
                gtk_widget_set_sensitive(w, TRUE);
	    w = pw->output_file_entry;
	    if(w != NULL)
		gtk_widget_set_sensitive(w, FALSE);
	    w = pw->output_file_browse_btn;
	    if(w != NULL)
                gtk_widget_set_sensitive(w, FALSE);
	}
	/* Output to file? */
	else if(widget == pw->output_file_radio)
	{
            w = pw->output_printer_entry;
            if(w != NULL)
                gtk_widget_set_sensitive(w, FALSE);
            w = pw->output_file_entry;
            if(w != NULL)
                gtk_widget_set_sensitive(w, TRUE);
            w = pw->output_file_browse_btn;
            if(w != NULL)
                gtk_widget_set_sensitive(w, TRUE);
	}

        /* Visual greyscale? */
        else if(widget == pw->visual_greyscale_radio)
        {
	    PrintWinUpdatePreviewVisual(pw);
            PrintWinPreviewDraw(pw);
        }
        /* Visual color? */
        else if(widget == pw->visual_color_radio)
        {
            PrintWinUpdatePreviewVisual(pw);
            PrintWinPreviewDraw(pw);
        }

        /* Orientation portrait? */
        else if(widget == pw->orientation_portrait_radio)
        {
	    gint tmp;

	    /* Need to swap output size. */
	    tmp = pw->output_width;
	    pw->output_width = pw->output_height;
	    pw->output_height = tmp;

	    PrintWinSyncOutputWidgets(pw);
            PrintWinRecreatePreview(pw);
            PrintWinPreviewDraw(pw);
        }
        /* Orientation landscape? */
        else if(widget == pw->orientation_landscape_radio)
        {
            gint tmp;

	    /* Need to swap output size. */
            tmp = pw->output_width;
            pw->output_width = pw->output_height;
            pw->output_height = tmp;

            PrintWinSyncOutputWidgets(pw);
            PrintWinRecreatePreview(pw);
	    PrintWinPreviewDraw(pw);
        }

	reenterent = FALSE;
}

/*
 *	Adjustment "value_changed" signal callback.
 */
static void PrintWinAdjustmentValueChangedCB(
        GtkAdjustment *adj, gpointer data
)
{
	static gbool reenterent = FALSE;
        print_win_struct *pw = (print_win_struct *)data;
        if((adj == NULL) || (pw == NULL))
            return;

	if(reenterent)
	    return;
	else
	    reenterent = TRUE;

	/* Assume brightness or contrast adjustment change. */
	PrintWinUpdatePreviewVisual(pw);

	/* Redraw. */
	if(pw->map_state)
	    PrintWinPreviewDraw(pw);

	reenterent = FALSE;
}


/*
 *	DPI entry "changed" signal callback.
 */
void PrintWinDPIChangedCB(GtkWidget *widget, gpointer data)
{
	static gbool reenterent = FALSE;
	const gchar *cstrptr;
        print_win_struct *pw = (print_win_struct *)data;
        if((widget == NULL) || (pw == NULL))
            return;

	if(reenterent)
	    return;
	else
	    reenterent = TRUE;

	/* Get text from entry widget and check if it contains
	 * data.
	 */
	cstrptr = gtk_entry_get_text(GTK_ENTRY(widget));
	if((cstrptr != NULL) ?
	    (((*cstrptr) != '\0') && !ISBLANK(*cstrptr)) : 0
	)
	{
	    /* The DPI has changed so the preview widget needs to be
	     * resized and the preview image needs to be recreated to
	     * match the new size of the preview widget.
	     */
	    PrintWinResizePreview(pw);
	}

	reenterent = FALSE;
}


/*
 *	Output geometry entry widget "changed" signal callback.
 */
static void PrintWinOutputEntryChangedCB(GtkWidget *widget, gpointer data)
{
	static gbool reenterent = FALSE;
	print_win_struct *pw = (print_win_struct *)data;
        if((widget == NULL) || (pw == NULL))
            return;

	/* Skip if manually updating entry widget. */
	if(pw->output_entry_update_manual)
	    return;

	if(reenterent)
	    return;
	else
	    reenterent = TRUE;

	if(widget == pw->output_x_entry)
	{
	    const gchar *cstrptr = gtk_entry_get_text(GTK_ENTRY(widget));
	    pw->output_x = (cstrptr != NULL) ? atoi(cstrptr) : 0;

            PrintWinPreviewDraw(pw);
	}
        if(widget == pw->output_y_entry)
        {
            const gchar *cstrptr = gtk_entry_get_text(GTK_ENTRY(widget));
            pw->output_y = (cstrptr != NULL) ? atoi(cstrptr) : 0;

            PrintWinPreviewDraw(pw);
        }
        if(widget == pw->output_width_entry)
        {
            const gchar *cstrptr = gtk_entry_get_text(GTK_ENTRY(widget));
            pw->output_width = (cstrptr != NULL) ? atoi(cstrptr) : 0;

	    if(pw->output_width > 0)
	    {
                PrintWinRecreatePreview(pw);
                PrintWinPreviewDraw(pw);
	    }
        }
        if(widget == pw->output_height_entry)
        {
            const gchar *cstrptr = gtk_entry_get_text(GTK_ENTRY(widget));
            pw->output_height = (cstrptr != NULL) ? atoi(cstrptr) : 0;

            if(pw->output_height > 0)
            {
                PrintWinRecreatePreview(pw);
                PrintWinPreviewDraw(pw);
	    }
        }


	reenterent = FALSE;
}

/*
 *	Print to file prompt browse button callback.
 */
static void PrintWinPrintFileBrowseCB(gpointer widget, gpointer data)
{
        GtkWidget *w;
        GtkEntry *entry;
        const gchar *last_path;
        fb_type_struct **ftype, *ftype_rtn;
        gint total_ftypes;
        gchar **path_rtn;
        gint path_total_rtns;
        print_win_struct *pw = (print_win_struct *)data;
        if(pw == NULL)
            return;

        w = pw->output_file_entry;
        if(w == NULL)
            return;

        entry = GTK_ENTRY(w);

        /* Get last path. */
        last_path = gtk_entry_get_text(entry);

        /* Set up file extension types list. */
        ftype = NULL;
        total_ftypes = 0;
        FileBrowserTypeListNew(
            &ftype, &total_ftypes,
            FTYPE_EXT_PS, "Adobe ISO PostScript"
        );
        FileBrowserTypeListNew(
            &ftype, &total_ftypes,
            FTYPE_EXT_PS2, "Adobe ISO PostScript 2"
        );
        FileBrowserTypeListNew(
            &ftype, &total_ftypes,
            FTYPE_EXT_PS3, "Adobe ISO PostScript 3"
        );
        FileBrowserTypeListNew(
            &ftype, &total_ftypes,
            "*.*", "All files"
        );

        /* Map file browser and query user for response. */
	FileBrowserSetTransientFor(pw->toplevel);
        if(FileBrowserGetResponse(
            "Select PostScript File", "Select", "Cancel",
            last_path,
            ftype, total_ftypes,
            &path_rtn, &path_total_rtns,
            &ftype_rtn
        ))
        {
            if(path_total_rtns > 0)
            {
                const gchar *new_path = (const gchar *)path_rtn[0];
                if(new_path != NULL)
                    gtk_entry_set_text(entry, new_path);
            }
        }
        FileBrowserSetTransientFor(NULL);

        /* Deallocate file extension types list. */
        FileBrowserDeleteTypeList(ftype, total_ftypes);
        ftype = NULL;
        total_ftypes = 0;
}


/*
 *	Print window's preview_da event handler.
 */
static gint PrintWinPreviewEventCB(
        GtkWidget *widget, GdkEvent *event, gpointer data
)
{
        gint etype, status = FALSE;
	gbool need_grab_pointer = FALSE;
        GdkCursor *cur = NULL;
        GdkWindow *window;
        GdkEventConfigure *configure;
	GdkEventCrossing *crossing;
        GdkEventKey *key;
        GdkEventButton *button;
        GdkEventMotion *motion;
        gbool keystate;
        guint keyval;
        gint x, y;
        GdkModifierType mask;
	imgview_image_struct *preview_img;
        print_win_struct *pw = (print_win_struct *)data;
        if((widget == NULL) || (event == NULL) || (pw == NULL))
            return(status);

        /* Get view widget GdkWindow. */
        if(GTK_WIDGET_NO_WINDOW(widget))
            window = NULL;
        else
            window = widget->window;

	/* Get preview image (may be NULL). */
	preview_img = pw->preview_img;

        /* Get event type. */
        etype = *(gint *)event;

        /* Handle by event type. */
        switch(etype)
        {
          case GDK_CONFIGURE:
            configure = (GdkEventConfigure *)event;
            PrintWinRecreatePreview(pw);
	    if(pw->map_state)
		PrintWinPreviewDraw(pw);
            status = TRUE;
            break;

          case GDK_ENTER_NOTIFY:
	    crossing = (GdkEventCrossing *)event;
	    if(!crossing->send_event)
	    {
		/* Need to reset all modifiers. */
/*
		if(pw->modifiers)
		{
		    pw->modifiers = 0;
                    if(window != NULL)
                        gdk_window_set_cursor(window, NULL);
		}
 */
		/* Set up widget to grab focus so "key_press_event" and
		 * and "key_release_event" events will be sent to it.
		 */
		gtk_widget_grab_focus(widget);

		/* Force tooltip to show, this will send a synthetic
		 * GDK_ENTER_NOTIFY event which we will ignore.
		 */
		GUIShowTipsNow(widget);
		status = TRUE;
	    }
            break;

          case GDK_LEAVE_NOTIFY:
            crossing = (GdkEventCrossing *)event;
            status = TRUE;
            break;

          case GDK_KEY_PRESS: case GDK_KEY_RELEASE:
            key = (GdkEventKey *)event;
            keyval = key->keyval;
            keystate = ((etype == GDK_KEY_PRESS) ? TRUE : FALSE);
            /* Handle by keyval. */
            /* ALT? */
            if((keyval == GDK_Alt_L) || (keyval == GDK_Alt_R))
            {
		if(keystate)
		    pw->modifiers |= PRINT_MODIFIER_ALT;
		else
		    pw->modifiers &= ~PRINT_MODIFIER_ALT;
                status = TRUE;
            }
            /* CTRL? */
            else if((keyval == GDK_Control_L) || (keyval == GDK_Control_R))
            {
                if(keystate)
                    pw->modifiers |= PRINT_MODIFIER_CTRL;
                else
                    pw->modifiers &= ~PRINT_MODIFIER_CTRL;
                status = TRUE;
            }
            /* SHIFT? */
            else if((keyval == GDK_Shift_L) || (keyval == GDK_Shift_R))
            {
                if(keystate)
                    pw->modifiers |= PRINT_MODIFIER_SHIFT;
                else
                    pw->modifiers &= ~PRINT_MODIFIER_SHIFT;
                status = TRUE;
            }
	    if(status)
	    {
		if(pw->drag_mode == PRINT_DRAG_NONE)
		{
		    if(pw->modifiers & PRINT_MODIFIER_SHIFT)
		    {
			cur = pw->size_cur;
		    }

                    /* Update cursor. */
                    if(window != NULL)
                        gdk_window_set_cursor(window, cur);
		}
	    }
	    break;

          case GDK_BUTTON_PRESS:
            button = (GdkEventButton *)event;
	    /* Handle by which button number was pressed. */
            switch(button->button)
            {
              case GDK_BUTTON1:
		if(pw->modifiers & (PRINT_MODIFIER_ALT | PRINT_MODIFIER_SHIFT))
		{
		    /* Size, same as if GDK_BUTTON2 was pressed. */
		    cur = pw->size_cur;
		    need_grab_pointer = TRUE;
                    pw->drag_mode = PRINT_DRAG_SIZE;
                    pw->drag_last_x = (gint)button->x;
                    pw->drag_last_y = (gint)button->y;

                    if(!PrintWinOutputOffsetW(
                        pw, &pw->drag_outline_x[0], &pw->drag_outline_y[0]
                    ))
                    {
                        pw->drag_outline_x[0] = 0;
                        pw->drag_outline_y[0] = 0;
                    }
                    if(preview_img != NULL)
                    {
                        pw->drag_outline_x[1] = pw->drag_outline_x[0] +
                            preview_img->width;
                        pw->drag_outline_y[1] = pw->drag_outline_y[0] +
                            preview_img->height;
                    }
                    /* If mapped, then draw only the outline. */
                    if(pw->map_state)
                    {
                        pw->have_prev_outline = FALSE;
                        pw->only_draw_outline = TRUE;;
                        PrintWinPreviewDraw(pw);
                    }
		}
		else
		{
		    /* No keyboard modifiers held, then just translate. */
                    cur = pw->translate_cur;
		    need_grab_pointer = TRUE;
		    pw->drag_mode = PRINT_DRAG_TRANSLATE;
		    pw->drag_last_x = (gint)button->x;
                    pw->drag_last_y = (gint)button->y;

		    if(!PrintWinOutputOffsetW(
		        pw, &pw->drag_outline_x[0], &pw->drag_outline_y[0]
		    ))
		    {
		        pw->drag_outline_x[0] = 0;
		        pw->drag_outline_y[0] = 0;
		    }
		    if(preview_img != NULL)
		    {
		        pw->drag_outline_x[1] = pw->drag_outline_x[0] +
			    preview_img->width;
                        pw->drag_outline_y[1] = pw->drag_outline_y[0] +
			    preview_img->height;
		    }
		    /* If mapped, then draw only the outline. */
                    if(pw->map_state)
		    {
		        pw->have_prev_outline = FALSE;
		        pw->only_draw_outline = TRUE;;
                        PrintWinPreviewDraw(pw);
		    }
		}
		status = TRUE;
		break;

              case GDK_BUTTON2:
		/* Size. */
                cur = pw->size_cur;
		need_grab_pointer = TRUE;
                pw->drag_mode = PRINT_DRAG_SIZE;
                pw->drag_last_x = (gint)button->x;
                pw->drag_last_y = (gint)button->y;

                if(!PrintWinOutputOffsetW(
                    pw, &pw->drag_outline_x[0], &pw->drag_outline_y[0]
                ))
                {
                    pw->drag_outline_x[0] = 0;
                    pw->drag_outline_y[0] = 0;
                }
                if(preview_img != NULL)
                {
                    pw->drag_outline_x[1] = pw->drag_outline_x[0] +
                        preview_img->width;
                    pw->drag_outline_y[1] = pw->drag_outline_y[0] +
                        preview_img->height;
                }
                /* If mapped, then draw only the outline. */
                if(pw->map_state)
                {
                    pw->have_prev_outline = FALSE;
                    pw->only_draw_outline = TRUE;;
                    PrintWinPreviewDraw(pw);
                }
                status = TRUE;
                break;
	    }
            /* Set window cursor if specified. */
            if((window != NULL) && (cur != NULL))
                gdk_window_set_cursor(window, cur);

            /* Grab pointer? */
            if(need_grab_pointer && (window != NULL))
            {
                gdk_flush();
                gdk_pointer_grab(
                    window,
                    FALSE,
                    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
                        GDK_POINTER_MOTION_MASK,
                    NULL,
                    GDK_NONE,
                    button->time
                );
		need_grab_pointer = FALSE;
                gdk_flush();
            }
            break;

          case GDK_BUTTON_RELEASE:
            button = (GdkEventButton *)event;

            /* Ungrab pointer. */
            gdk_flush();
            gdk_pointer_ungrab(button->time);
            gdk_flush();
            while(gdk_pointer_is_grabbed())
                { gdk_pointer_ungrab(GDK_CURRENT_TIME); gdk_flush(); }

	    /* Handle by current drag mode. */
            switch(pw->drag_mode)
            {
              case PRINT_DRAG_TRANSLATE:
		if(1)
		{
		    gint	ox = MIN(
			pw->drag_outline_x[0], pw->drag_outline_x[1]
		    ),
				oy = MIN(
			pw->drag_outline_y[0], pw->drag_outline_y[1]
		    );
		    gint paper_width, paper_height, ww, wh;
		    gfloat wtoo_coeff;


		    ww = widget->allocation.width;
		    wh = widget->allocation.height;

		    if(!PrintWinPaperSize(pw, &paper_width, &paper_height, TRUE))
		    {
			paper_width = 320;
			paper_height = 240;
		    }

		    if(ww > 0)
			wtoo_coeff = (gfloat)paper_width / (gfloat)ww;
		    else
			wtoo_coeff = 0.0;
		    pw->output_x = ox * wtoo_coeff;

                    if(wh > 0)
                        wtoo_coeff = (gfloat)paper_height / (gfloat)wh;
                    else
                        wtoo_coeff = 0.0;
                    pw->output_y = oy * wtoo_coeff;

		    /* Update output entry widgets to reflect new position. */
		    PrintWinSyncOutputWidgets(pw);
		}
                /* Reset drag mode and redraw. */
                pw->drag_mode = PRINT_DRAG_NONE;
                if(pw->map_state)
                    PrintWinPreviewDraw(pw);
                status = TRUE;
                break;

              case PRINT_DRAG_SIZE:
                if(1)
                {
		    gint	ow = ABSOLUTE(
			pw->drag_outline_x[1] - pw->drag_outline_x[0]
		    ),
				oh = ABSOLUTE(
			pw->drag_outline_y[1] - pw->drag_outline_y[0]
                    );
                    gint paper_width, paper_height, ww, wh;
                    gfloat wtoo_coeff;


                    ww = widget->allocation.width;
                    wh = widget->allocation.height;

		    if(!PrintWinPaperSize(pw, &paper_width, &paper_height, TRUE))
                    {
                        paper_width = 320;
                        paper_height = 240;
                    }

                    if(ww > 0)
                        wtoo_coeff = (gfloat)paper_width / (gfloat)ww;
                    else
                        wtoo_coeff = 0.0;
		    if((ow > 0) && (wtoo_coeff > 0.0))
			pw->output_width = ow * wtoo_coeff;

                    if(wh > 0)
                        wtoo_coeff = (gfloat)paper_height / (gfloat)wh;
                    else
                        wtoo_coeff = 0.0;
                    if((oh > 0) && (wtoo_coeff > 0.0))
                        pw->output_height = oh * wtoo_coeff;

                    /* Update output entry widgets to reflect new size. */
                    PrintWinSyncOutputWidgets(pw);

		    /* Need to recreate the preview image since it changed
		     * size.
		     */
		    PrintWinRecreatePreview(pw);
                }
		/* Reset drag mode and redraw. */
                pw->drag_mode = PRINT_DRAG_NONE;
                if(pw->map_state)
                    PrintWinPreviewDraw(pw);
                status = TRUE;
                break;

	      default:
		/* Always reset drag mode on button release. */
		pw->drag_mode = PRINT_DRAG_NONE;
		break;
            }
	    /* If no keyboard modifiers are being held then reset cursor. */
	    if(!pw->modifiers)
	    {
                if(window != NULL)
                    gdk_window_set_cursor(window, NULL);
	    }
            break;

          case GDK_MOTION_NOTIFY:
            motion = (GdkEventMotion *)event;
            if(motion->is_hint)
            {
                if(pw->drag_mode != PRINT_DRAG_NONE)
                    gdk_window_get_pointer(
                        motion->window, &x, &y, &mask
                    );
            }
            else
            {
                x = (gint)motion->x;
                y = (gint)motion->y;
                mask = motion->state;
            }

            /* Handle by drag mode. */
            switch(pw->drag_mode)
            {
              case PRINT_DRAG_TRANSLATE:
		if(1)
		{
		    gint	dx = x - pw->drag_last_x,
				dy = y - pw->drag_last_y;

		    pw->drag_outline_x[0] += dx;
		    pw->drag_outline_y[0] += dy;
                    pw->drag_outline_x[1] += dx;
                    pw->drag_outline_y[1] += dy;
		}
                /* If mapped, then draw only the outline. */
                if(pw->map_state)
                {
                    pw->only_draw_outline = TRUE;;
                    PrintWinPreviewDraw(pw);
                }
                /* Update last drag positions. */
                pw->drag_last_x = (gint)x;
                pw->drag_last_y = (gint)y;
                status = TRUE;
		break;

	      case PRINT_DRAG_SIZE:
                if(1)
                {
                    gint        dx = x - pw->drag_last_x,
                                dy = y - pw->drag_last_y;

                    pw->drag_outline_x[1] += dx;
                    pw->drag_outline_y[1] += dy;
                }
                /* If mapped, then draw only the outline. */
                if(pw->map_state)
                {
                    pw->only_draw_outline = TRUE;;
                    PrintWinPreviewDraw(pw);
                }
		/* Update last drag positions. */
                pw->drag_last_x = (gint)x;
                pw->drag_last_y = (gint)y;
                status = TRUE;
		break;
	    }
	    break;
        }

        return(status);
}

/*
 *	Print window's preview_da widget "expose_event" callback.
 */
static gint PrintWinPreviewExposeCB(
        GtkWidget *widget, GdkEventExpose *expose, gpointer data
)
{
	GdkGC *gc_base, *gc_invert;
	gbool only_draw_outline;
	GdkWindow *window;
	GdkDrawable *d;
	GtkWidget *w;
	gint ww, wh;
	imgview_image_struct *img;
	print_win_struct *pw = (print_win_struct *)data;
	if((widget == NULL) || (pw == NULL))
	    return(FALSE);

	/* Get marker for if we only want to draw the drag outlines. */
	only_draw_outline = pw->only_draw_outline;
	pw->only_draw_outline = FALSE;	/* Always reset it in this function. */

	/* Get preview drawing area widget. */
	w = pw->preview_da;
        if(w != widget)
            return(FALSE);

	ww = w->allocation.width;
	wh = w->allocation.height;
	if((ww < 1) || (wh < 1))
	    return(FALSE);

	/* Get pointers to the GCs. */
	gc_base = pw->gc_base;
	gc_invert = pw->gc_invert;

	/* Must have the base GC. */
	if(gc_base == NULL)
	    return(FALSE);

	/* Get preview drawing area widget's window. */
	if(GTK_WIDGET_NO_WINDOW(w))
	    window = NULL;
	else
	    window = w->window;
	if(window == NULL)
	    return(FALSE);

	d = (GdkDrawable *)window;


	/* Clear background as needed. */
	if(!only_draw_outline)
	{
	    gdk_draw_rectangle(d, gc_base, TRUE, 0, 0, ww, wh);
	}

	/* Get preview image and check if it contains data? */
	img = pw->preview_img;
	if((img != NULL) ?
	    ((img->mem != NULL) && !only_draw_outline) : 0
	)
	{
	    gint x, y;


	    if(!PrintWinOutputOffsetW(pw, &x, &y))
	    {
		x = 0;
		y = 0;
	    }

	    switch(img->bpp)
	    {
	      case 4:
		gdk_draw_rgb_32_image(
		    d, gc_base,
		    x, y,
		    img->width, img->height,
		    GDK_RGB_DITHER_NORMAL,
		    img->mem,
		    img->bpl
		);
		break;

	      case 3:
		gdk_draw_rgb_image(
		    d, gc_base,
		    x, y,
                    img->width, img->height,
                    GDK_RGB_DITHER_NORMAL,
                    img->mem,
                    img->bpl
                );
                break;
	    }
	}


	/* Clear previous drag marks if defined and if only drawing
	 * for drag outlines.
	 */
	if(pw->have_prev_outline && only_draw_outline && (gc_invert != NULL))
	{
	     gint        x = MIN(pw->prev_outline_x[0], pw->prev_outline_x[1]),
                         y = MIN(pw->prev_outline_y[0], pw->prev_outline_y[1]);
             gint        width = ABSOLUTE(
                pw->prev_outline_x[1] - pw->prev_outline_x[0]
                         ),
                         height = ABSOLUTE(
                pw->prev_outline_y[1] - pw->prev_outline_y[0]
                         );

	    if((width > 0) && (height > 0))
                gdk_draw_rectangle(
                    d, gc_invert, FALSE, x, y, width, height
                );

	    pw->have_prev_outline = FALSE;
	}

	/* Draw new drag marks? */
	if((pw->drag_mode != PRINT_DRAG_NONE) && (gc_invert != NULL))
	{
	    gint	x = MIN(pw->drag_outline_x[0], pw->drag_outline_x[1]),
			y = MIN(pw->drag_outline_y[0], pw->drag_outline_y[1]);
	    gint	width = ABSOLUTE(
		pw->drag_outline_x[1] - pw->drag_outline_x[0]
			),
			height = ABSOLUTE(
		pw->drag_outline_y[1] - pw->drag_outline_y[0]
			);

	    if((width > 0) && (height > 0))
	    {
		gdk_draw_rectangle(
		    d, gc_invert, FALSE, x, y, width, height
		);

		/* Record newly drawn drag outline marks. */
		pw->have_prev_outline = TRUE;
		pw->prev_outline_x[0] = pw->drag_outline_x[0];
                pw->prev_outline_y[0] = pw->drag_outline_y[0];
                pw->prev_outline_x[1] = pw->drag_outline_x[1];
                pw->prev_outline_y[1] = pw->drag_outline_y[1];
	    }
	}

	return(TRUE);
}


/*
 *	Sets the output entry widgets on the print window to match
 *	the current output geometry values.
 */
static void PrintWinSyncOutputWidgets(print_win_struct *pw)
{
	GtkWidget *w;
	gchar num_str[80];


	if(pw == NULL)
	    return;

	/* Mark that we are manually updating the output entry widgets,
	 * this is so that the "changed" signal will be ignored.
	 */
	pw->output_entry_update_manual = TRUE;


	/* Begin updating output entry widgets. */
	sprintf(num_str, "%i", pw->output_x);
	w = pw->output_x_entry;
	if(w != NULL)
	    gtk_entry_set_text(GTK_ENTRY(w), num_str);

        sprintf(num_str, "%i", pw->output_y);
        w = pw->output_y_entry;
        if(w != NULL)
            gtk_entry_set_text(GTK_ENTRY(w), num_str);

        sprintf(num_str, "%i", pw->output_width);
        w = pw->output_width_entry;
        if(w != NULL)
            gtk_entry_set_text(GTK_ENTRY(w), num_str);

        sprintf(num_str, "%i", pw->output_height);
        w = pw->output_height_entry;
        if(w != NULL)
            gtk_entry_set_text(GTK_ENTRY(w), num_str);

	/* Mark that we are no longer manually updating the output
	 * entry widgets.
	 */
	pw->output_entry_update_manual = FALSE;
}

/*
 *	Gets the visual, one of PRINT_VISUAL_*.
 */
static gint PrintWinVisual(print_win_struct *pw)
{
        GtkWidget *w;

        if(pw == NULL)
            return(PRINT_VISUAL_GREYSCALE);

        w = pw->visual_greyscale_radio;
        if((w != NULL) ? !GTK_IS_TOGGLE_BUTTON(w) : 1)
            return(PRINT_VISUAL_GREYSCALE);

        if(GTK_TOGGLE_BUTTON(w)->active)
            return(PRINT_VISUAL_GREYSCALE);
        else
            return(PRINT_VISUAL_COLOR);
}

/*
 *	Gets the orientation, one of PRINT_ORIENTATION_*.
 */
static gint PrintWinOrientation(print_win_struct *pw)
{
	GtkWidget *w;

	if(pw == NULL)
	    return(PRINT_ORIENTATION_PORTRAIT);

	w = pw->orientation_portrait_radio;
	if((w != NULL) ? !GTK_IS_TOGGLE_BUTTON(w) : 1)
	    return(PRINT_ORIENTATION_PORTRAIT);

	if(GTK_TOGGLE_BUTTON(w)->active)
	    return(PRINT_ORIENTATION_PORTRAIT);
	else
	    return(PRINT_ORIENTATION_LANDSCAPE);
}

/*
 *	Gets teh DPI (dots per inch) value, returns TRUE if the value
 *	was obtained or FALSE on failure.
 */
static gbool PrintWinGetDPI(
        print_win_struct *pw,
        gint *dpi_x, gint *dpi_y
)
{
        GtkWidget *w;

        if(dpi_x != NULL)
            (*dpi_x) = -1;
        if(dpi_y != NULL)
            (*dpi_y) = -1;

        if(pw == NULL)
            return(FALSE);


	w = pw->dpi_x_entry;
	if((w != NULL) ? GTK_IS_ENTRY(w) : 0)
	{
	    const gchar *cstrptr = gtk_entry_get_text(
		GTK_ENTRY(w)
	    );
	    if((cstrptr != NULL) && (dpi_x != NULL))
		(*dpi_x) = atoi(cstrptr);
	}

        w = pw->dpi_y_entry;
        if((w != NULL) ? GTK_IS_ENTRY(w) : 0)
        {
            const gchar *cstrptr = gtk_entry_get_text(
                GTK_ENTRY(w)
            );
            if((cstrptr != NULL) && (dpi_y != NULL))
                (*dpi_y) = atoi(cstrptr);
        }

	return(TRUE);
}

/*
 *	Gets the paper size, returns TRUE if the value was obtained or
 *	FALSE on failure.
 *
 *	If apply_dpi is TRUE then the dpi will be applied to the paper
 *	size before it's returned.
 */
static gbool PrintWinPaperSize(
        print_win_struct *pw,
        gint *width, gint *height,
        gbool apply_dpi
)
{
	GtkCList *clist;
	GList *glist;
	gint row;
	const gchar **strv;
	gint orig_width, orig_height;


	if(width != NULL)
	    (*width) = -1;
	if(height != NULL)
	    (*height) = -1;

        if(pw == NULL)
            return(FALSE);

	clist = (GtkCList *)pw->paper_size_clist;
	if(clist == NULL)
	    return(FALSE);

	/* Get last selected row. */
	glist = clist->selection;
	row = -1;
	while(glist != NULL)
	{
	    row = (gint)glist->data;
	    glist = glist->next;
	}

	/* Got valid row? */
	if((row < 0) || (row >= clist->rows))
	    return(FALSE);

	/* Get row data, an array of 4 string pointers.
	 * The pointer index 0 and 1 point to allocated strings,
	 * while 2 and 3 point to gint's of width and height.
	 */
	strv = (const gchar **)gtk_clist_get_row_data(clist, row);
	if(strv == NULL)
	    return(FALSE);

	/* Get width and height from row data, these values are in
	 * pixels and assumed dpi of 100.
	 */
	orig_width = (gint)strv[2];
	orig_height = (gint)strv[3];

	if((orig_width <= 0) || (orig_height <= 0))
	    return(FALSE);

	/* Add DPI to dimensions? */
	if(apply_dpi)
	{
	    gint dpi_x, dpi_y;

	    /* Get DPI. */
	    if(PrintWinGetDPI(pw, &dpi_x, &dpi_y))
	    {
		/* Modify width and height to match DPI. */
		orig_width = (gfloat)orig_width *
		    ((gfloat)dpi_x / 100.0);
                orig_height = (gfloat)orig_height *
                    ((gfloat)dpi_y / 100.0);
	    }
	}

	/* Update returns. */
	if(width != NULL)
            (*width) = orig_width;
        if(height != NULL)
            (*height) = orig_height;

	return(TRUE);
}

/*
 *	Returns the output image offset to its upper left corner
 *	from the upper left corner (regardless of orientation) in
 *	window coordinates.
 *
 *	Returns TRUE if the output x and y offsets were obtained
 *	or FALSE on error.
 */
static gbool PrintWinOutputOffsetW(
	print_win_struct *pw, gint *x, gint *y
)
{
	gfloat otow_coeff;
	gint paper_width, paper_height;
	GtkWidget *w;


	if(pw == NULL)
	    return(FALSE);

	w = pw->preview_da;
	if(w == NULL)
	    return(FALSE);

	if(!PrintWinPaperSize(
	    pw, &paper_width, &paper_height, TRUE
	))
	    return(FALSE);

	if((paper_width < 1) || (paper_height < 1))
	{
	    if(x != NULL)
		(*x) = 0;
            if(y != NULL)
                (*y) = 0;

	    return(TRUE);
	}

	/* Calculate offsets. */
	otow_coeff = (gfloat)w->allocation.width / (gfloat)paper_width;
	if(x != NULL)
	    (*x) = pw->output_x * otow_coeff;
	otow_coeff = (gfloat)w->allocation.height / (gfloat)paper_height;
	if(y != NULL)
	    (*y) = pw->output_y * otow_coeff;

	return(TRUE);
}

/*
 *	Coppies the image data from src_data to the target image
 *	tar_data who's dimensions are rotated 90 degrees counter
 *	clockwise.
 */
static void PrintWinLandscapeImage(
        gint bpp,
        const guint8 *src_data,
        gint swidth, gint sheight, gint sbpl,
        guint8 *tar_data,
        gint tbpl
)
{
	gint sx, sy, tx, ty, bc;
	const guint8 *src_ptr;
	guint8 *tar_ptr;
	gint twidth, theight;


	if((tar_data == NULL) || (src_data == NULL) || (bpp < 1))
	    return;

	if((swidth < 1) || (sheight < 1) || (sbpl < 1))
	    return;

	/* Get target image buffer size which should be the size of the
	 * source image rotated 90 degrees.
	 */
	twidth = sheight;
	theight = swidth;


	/* Itterate through source image. */
	for(sy = 0; sy < sheight; sy++)
	{
	    for(sx = 0; sx < swidth; sx++)
	    {
		tx = sy;
		ty = theight - 1 - sx;

		src_ptr = &src_data[
		    (sy * sbpl) + (sx * bpp)
		];
                tar_ptr = &tar_data[
                    (ty * tbpl) + (tx * bpp)
                ];

		for(bc = 0; bc < bpp; bc++)
		    *tar_ptr++ = *src_ptr++;
	    }
	}
}


/*
 *	Coppies image data from src_img to tar_img. Both images must have
 *	the same geometry and bpp.
 */
static void PrintWinRevertImages(
        imgview_image_struct *src_img, imgview_image_struct *tar_img
)
{
	gint y, bpp, src_bpl, tar_bpl, min_bpl, height;
	const guint8 *src_data, *src_ptr;
	guint8 *tar_data, *tar_ptr;


	if((src_img == NULL) || (tar_img == NULL))
	    return;

	/* Source image geometry and bpp must match the target image. */
	if((src_img->bpp != tar_img->bpp) ||
           (src_img->width != tar_img->width) ||
	   (src_img->height != tar_img->height)
	)
	{
	    fprintf(
		stderr,
"PrintWinRevertImages(): BPP or geometry mismatch, cannot copy.\n"
	    );
	    return;
	}

	/* Get bpp and height from source image, they should be the same
	 * on the target image.
	 */
	bpp = src_img->bpp;
	height = src_img->height;

	/* Get bytes per line for each image, they may be different. */
	src_bpl = src_img->bpl;
	tar_bpl = tar_img->bpl;

	/* Calculate smaller of the two image's bytes per line values. */
	min_bpl = MIN(src_bpl, tar_bpl);
	if(min_bpl < 1)
	    return;

	/* Get pointers to both image's buffers, both must be allocated. */
	src_data = (const guint8 *)src_img->mem;
	tar_data = (guint8 *)tar_img->mem;
	if((src_data == NULL) || (tar_data == NULL))
	    return;

	/* Itterate through each line (note height will be the same on
	 * both images). Copy each line.
	 */
	for(y = 0; y < height; y++)
	{
	    /* Get pointer to current source and target lines. */
	    src_ptr = &src_data[y * src_bpl];
	    tar_ptr = &tar_data[y * tar_bpl];

	    /* Copy the common amount of bytes on current line. */
	    memcpy(tar_ptr, src_ptr, min_bpl);
	}
}


/*
 *	Resets the image data in preview_img with the data from 
 *	orig_preview_img.
 */
static void PrintWinRevertPreview(print_win_struct *pw)
{
	if(pw == NULL)
	    return;

        /* Copy data from the original preview image to the preview
         * image that will be ready for drawing on preview_da.
         */
	PrintWinRevertImages(
	    pw->orig_preview_img,	/* Source. */
	    pw->preview_img		/* Target. */
	);
}


/*
 *	Modifies the given image with respect to the given
 *	print values.
 */
static void PrintWinUpdatePreviewVisualV(
        imgview_image_struct *img, print_values_struct *v
)
{
	guint vflags;
        gint visual = PRINT_VISUAL_GREYSCALE;
        gfloat brightness = 0.0, contrast = 0.0;
        gint x, y, bc;
        gint bpp, min_bpp, bpl, width, height;
        guint8 *data, *data_ptr;
        guint32 cur_byte, avg_byte;
        gint brightness_delta;


	if((img == NULL) || (v == NULL))
	    return;

	vflags = v->flags;


	/* Get values from image. */
        data = img->mem;
        if(data == NULL)
            return;
        bpp = img->bpp;
        bpl = img->bpl;
        width = img->width;
        height = img->height;
        /* Calculate minimized bytes per pixel to be 3 or less since
         * we are never interested in the fourth alpha byte.
         */
        min_bpp = MIN(bpp, 3);

	/* Get print values. */
	if(vflags & PRINT_VALUE_VISUAL)
	    visual = v->visual;
        if(vflags & PRINT_VALUE_BRIGHTNESS)
            brightness = v->brightness;
        if(vflags & PRINT_VALUE_CONTRAST)
            contrast = v->contrast;


        /* Calculate brightness delta byte values. */
        brightness_delta = (gint)(0xff * brightness);

        /* Itterate through image pixels. */
        for(y = 0; y < height; y++)
        {
            for(x = 0; x < width; x++)
            {
                data_ptr = &data[
                    (y * bpl) + (x * bpp)
                ];

                /* Begin per pixel operations. */

                /* Calculate the average value (we'll need it in several
                 * places. The average byte value is from 0x00 to 0xff.
                 */
		avg_byte = (gint)(
		    (0.299 * data_ptr[0]) + (0.587 * data_ptr[1]) +
		    (0.114 * data_ptr[2])
		);

                /* Make pixel grey? */
                if(visual == PRINT_VISUAL_GREYSCALE)
                {
                    /* Set averaged value for each byte of the pixel. */
                    for(bc = 0; bc < min_bpp; bc++)
                        data_ptr[bc] = avg_byte;
                }

                /* Adjust brightness? */
                if(brightness_delta != 0)
                {
                    /* Itterate through non-alpha color bytes of current
                     * pixel.
                     */
                    for(bc = 0; bc < min_bpp; bc++)
                        data_ptr[bc] = (guint8)CLIP(
                            (gint)data_ptr[bc] + brightness_delta,
                            0x00, 0xff
                        );
                }
                /* Adjust contrast? */
                if(contrast != 0.0)
                {
                    /* Itterate through non-alpha color bytes of current
                     * pixel.
                     */
                    for(bc = 0; bc < min_bpp; bc++)
                    {
                        cur_byte = (guint32)(data_ptr[bc]);
                        data_ptr[bc] = (guint8)CLIP(
                            (gint)cur_byte +
			    (gint)(((gint)cur_byte - (gint)0x80) * contrast),
			    0x00, 0xff
                        );
                    }
                }

            }
        }
}


/*
 *	Updates the preview_img to reflect the new values of the
 *	current visual values.
 *
 *	Note that PrintWinRevertPreview() will be called to revert the
 *	data first.
 */
static void PrintWinUpdatePreviewVisual(print_win_struct *pw)
{
	GtkAdjustment *adj;
	print_values_struct v;


	if(pw == NULL)
	    return;

	/* Copy data from orig_preview_img to preview_img. */
	PrintWinRevertPreview(pw);

	/* Set up print values. */
	v.flags = (PRINT_VALUE_BRIGHTNESS | PRINT_VALUE_CONTRAST |
	    PRINT_VALUE_VISUAL
	);
        adj = pw->brightness_adj;
        v.brightness = ((adj != NULL) ? adj->value : 0.0);
        adj = pw->contrast_adj;
        v.contrast = ((adj != NULL) ? adj->value : 0.0);
	v.visual = PrintWinVisual(pw);

	/* Modify preview_img's visual, brightness, and contrast. */
	PrintWinUpdatePreviewVisualV(pw->preview_img, &v);
}

/*
 *	Recreates the original preview image orig_preview_img and
 *	the preview image preview_img. The old images will be destroyed
 *	if they exist.
 *
 *	The new preview_img will contain the same data as 
 *	orig_preview_img.
 *
 *	This function will call PrintWinUpdatePreviewVisual() to make sure
 *	the new images are updated and ready to be drawn.
 */
static void PrintWinRecreatePreview(print_win_struct *pw)
{
	gint orientation, visual, depth;
	gint pwidth, pheight;
	gint output_width, output_height;
	gint new_width, new_height;
	gfloat x_coeff, y_coeff;
	GdkWindow *window;
	GtkWidget *preview_da;
	imgview_image_struct *tar_img;


	if(pw == NULL)
	    return;

	/* Destroy old preview images. */
	if((pw->preview_img != NULL) || (pw->orig_preview_img != NULL))
	{
	    gdk_flush();
	    ImgViewImageDelete(pw->preview_img);
	    pw->preview_img = NULL;
            ImgViewImageDelete(pw->orig_preview_img);
            pw->orig_preview_img = NULL;
	}


	/* Get preview GtkDrawingArea widget. */
	preview_da = pw->preview_da;
	if(preview_da == NULL)
	    return;

	/* Get orientation and visual. */
	orientation = PrintWinOrientation(pw);
	visual = PrintWinVisual(pw);

	/* Get new paper size. */
	if(!PrintWinPaperSize(pw, &pwidth, &pheight, TRUE))
	    return;

	/* Get output image size in pixels, we can ignore the
	 * the orientation because the output size should already
	 * be updated for the current orientation.
	 */
	output_width = pw->output_width;
	output_height = pw->output_height;
	if((output_width <= 0) || (output_height <= 0))
	    return;

	/* Calculate coefficients that represent the amount of
	 * the output image visible on the paper.
	 */
	x_coeff = (gfloat)output_width / (gfloat)pwidth;
        y_coeff = (gfloat)output_height / (gfloat)pheight;

	/* Get window from preview widget. */
	window = preview_da->window;
	depth = 0;
	if(GTK_WIDGET_NO_WINDOW(preview_da) ? 0 : (window != NULL))
	{
	    GdkVisual *vis = gdk_window_get_visual(window);
	    if(vis != NULL)
		depth = vis->depth;
	}

	/* Calculate new preview image size. */
	new_width = (gint)(preview_da->allocation.width * x_coeff);
	new_height = (gint)(preview_da->allocation.height * y_coeff);

	if((new_width < 1) || (new_height < 1))
	    return;


	/* Create new preview images. */
	pw->orig_preview_img = tar_img = ImgViewImageNew(
	    new_width, new_height, depth
	);
        pw->preview_img = ImgViewImageNew(
            new_width, new_height, depth
        );

	/* Original preview image created? If so then we update
	 * it to contain the data from the image viewer with some
	 * modifications. This image will be resized to match the preview
	 * widget's window coordinates, rotated, and color adjusted.
	 */
	if(tar_img != NULL)
	{
	    /* Update new preview image's data. */
	    guint8 *tar_data = tar_img->mem;
	    const guint8 *src_data;
	    gint swidth, sheight, sbpp, sbpl, sformat;


	    /* Get source image attributes. */
	    src_data = (const guint8 *)ImgViewGetImageData(
		pw->iv, &swidth, &sheight, &sbpl, &sbpp, &sformat
	    );
	    if(src_data != NULL)
	    {
		if(sbpp != tar_img->bpp)
		{
		    fprintf(
			stderr,
"PrintWinRecreatePreview(): BPP mismatch on source and target images.\n"
		    );
		    return;
		}

		/* Copy by orientation. */
		if(orientation == PRINT_ORIENTATION_LANDSCAPE)
		{
		    /* Create a tempory image buffer. */
		    guint8 *tmp_data;
		    gint twidth, theight, tbpl;


		    /* Get size of tempory image buffer, the size should
		     * match the preview image's size but be rotated 90
		     * degrees so width and height are swapped.
		     */
		    twidth = tar_img->height;	/* Swap width and height. */
		    theight = tar_img->width;
		    tbpl = twidth * sbpp;
		    tmp_data = (guint8 *)g_malloc(theight * tbpl);
		    if(tmp_data != NULL)
		    {
			/* Copy source image buffer to tempory image
			 * buffer and resize.
			 */
                        GUIResizeBuffer(
                            sbpp,
                            (const gpointer)src_data,
                            swidth, sheight, sbpl,
                            (gpointer)tmp_data,
                            twidth, theight, tbpl
                        );

			/* Now copy and rotate tempory buffer to new
			 * preview image buffer. Note that the target
			 * image's width and height will be derived from
			 * the swapped size of the source image tmp_data.
			 *
			 * From above twidth and theight are calculated
			 * from tar_img->height and tar_img->width 
			 * (respectivly).
			 */
			PrintWinLandscapeImage(
                            sbpp,
                            (const gpointer)tmp_data,
                            twidth, theight, tbpl,
                            (gpointer)tar_data,
                            tar_img->bpl
			);

			/* Deallocate tempory image buffer, it is no
			 * longer needed.
			 */
			g_free(tmp_data);
			tmp_data = NULL;
		    }
		}
		else
		{
		    /* Resize and copy source image to preview image. */
		    GUIResizeBuffer(
		        sbpp,
		        (const gpointer)src_data,
		        swidth, sheight, sbpl,
		        (gpointer)tar_data,
		        tar_img->width, tar_img->height, tar_img->bpl
		    );
		}
	    }
	}

	/* This will call PrintWinRevertPreview() to first copy the data
	 * from orig_preview_img to preview_img and modify preview_img to
	 * reflect the current visual settings.
	 */
	PrintWinUpdatePreviewVisual(pw);
}

/*
 *	Resizes the preview drawing area widget.
 */
static void PrintWinResizePreview(print_win_struct *pw)
{
	GtkWidget *w;
	gint width, height;	/* Nominal size of preview_da. */
	gint new_width, new_height;
        gint pwidth, pheight;
	gfloat aspect;


        if(pw == NULL)
            return;

	w = pw->preview_da;
	if(w == NULL)
	    return;

	/* Set nominal display size of preview drawing area. */
	width = PRINT_PREVIEW_WIDTH;
	height = PRINT_PREVIEW_HEIGHT;

        /* Get new paper size. */
        if(!PrintWinPaperSize(pw, &pwidth, &pheight, TRUE))
            return;
	if(pheight <= 0)
	    return;

	/* Calculate aspect ratio. */
	aspect = (gfloat)pwidth / (gfloat)pheight;
	if(aspect <= 0.0)
	    return;

	/* Calculate new size for preview drawing area. */
	if(aspect >= 1.0)
	{
	    /* Width is longer than height. */
	    new_width = width;
	    new_height = height / aspect;
	}
	else
	{
	    /* Height is longer than width. */
	    new_width = width * aspect;
	    new_height = height;
	}

	/* Set new size of preview drawing area. */
	w = pw->preview_da;
	if(w != NULL)
	    gtk_widget_set_usize(w, new_width, new_height);

	/* Force toplevel to resize. */
	w = pw->toplevel;
	if(w != NULL)
	    gtk_widget_queue_resize(w);

	/* We don't need to recreate the preview images since a
	 * "configure_event" will be generated and the signal handler
	 * will see it and recreate the preview images.
	 */
}


/*
 *	This basically calls PrintWinPreviewExposeCB().
 */
static void PrintWinPreviewDraw(print_win_struct *pw)
{
	if(pw == NULL)
	    return;

	PrintWinPreviewExposeCB(pw->preview_da, NULL, pw);
}



/*
 *	Creates a new print window.
 */
print_win_struct *PrintWinNew(imgview_struct *iv)
{
	gint	bw = (100 + (2 * 3)),
		bh = (30 + (2 * 3)),
		border_major = 5,
		border_minor = 2;
	gint i;
	GSList *gslist;
	GdkColormap *colormap;
	GdkWindow *window;
	GtkStyle *style;
	GtkAccelGroup *accelgrp;
	GtkAdjustment *adj;
	GtkWidget	*w, *parent, *parent2, *parent3, *parent4,
			*parent5, *parent6, *main_vbox, *scroll_parent;
	gpointer entry, browse_btn;
	gchar *heading[4];
	const gchar *paper_size[] = PAPER_SIZE_LIST;
	print_win_struct *pw = (print_win_struct *)g_malloc0(
	    sizeof(print_win_struct)
	);
	if(pw == NULL)
	    return(NULL);

gdk_rgb_init();

	/* Begin resetting values. */
	pw->initialized = TRUE;
        pw->map_state = FALSE;
        pw->iv = iv;
	pw->preview_img = NULL;
	pw->orig_preview_img = NULL;
	pw->drag_mode = PRINT_DRAG_NONE;


	/* Load pointer cursors. */
	pw->busy_cur = gdk_cursor_new(GDK_WATCH);
	pw->translate_cur = gdk_cursor_new(GDK_FLEUR);
	pw->size_cur = gdk_cursor_new(GDK_SIZING);

        /* Keyboard accelerator group. */
        pw->accelgrp = accelgrp = gtk_accel_group_new();

        /* Toplevel. */
        gtk_widget_push_visual(gdk_rgb_get_visual());
        gtk_widget_push_colormap(gdk_rgb_get_cmap());
        pw->toplevel = w = gtk_window_new(GTK_WINDOW_DIALOG);
        gtk_widget_pop_visual();
        gtk_widget_pop_colormap();
        gtk_window_set_policy(
            GTK_WINDOW(w), TRUE, TRUE, TRUE
        );
	gtk_widget_set_usize(w, PRINT_DEF_WIDTH, PRINT_DEF_HEIGHT);
        gtk_widget_realize(w);
        gtk_window_set_title(GTK_WINDOW(w), PRINT_DEF_TITLE);
        window = w->window;
        if(window != NULL)
        {
            gdk_window_set_decorations(
                window,
                GDK_DECOR_TITLE | GDK_DECOR_MENU | GDK_DECOR_MINIMIZE
            );
            gdk_window_set_functions(
                window,
                GDK_FUNC_MOVE | GDK_FUNC_MINIMIZE | GDK_FUNC_CLOSE
            );
	    GUISetWMIcon(window, (u_int8_t **)icon_print_48x48_xpm);
        }
        gtk_signal_connect(
            GTK_OBJECT(w), "delete_event",
            GTK_SIGNAL_FUNC(PrintWinCloseCB),
            (gpointer)pw
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "destroy",
            GTK_SIGNAL_FUNC(PrintWinDestroyCB),
            (gpointer)pw
        );
        gtk_container_border_width(GTK_CONTAINER(w), 0);
        gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
        parent = w;


	/* Main vbox. */
        main_vbox = w = gtk_vbox_new(FALSE, 0);
        gtk_container_add(GTK_CONTAINER(parent), w);
        gtk_widget_show(w);
        parent = w;

	/* Vbox for child widgets. */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(main_vbox), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
        parent = w;


	/* Output widgets. */
	w = gtk_frame_new("Print To");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
        parent2 = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent2 = w;

	/* Hbox for print command widgets. */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
        parent3 = w;

	/* Print to printer radio. */
	gslist = NULL;
        pw->output_printer_radio = w =
	    gtk_radio_button_new_with_label(gslist, "Printer");
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "toggled",
            GTK_SIGNAL_FUNC(PrintWinToggledCB),
            (gpointer)pw
        );
        gtk_widget_set_usize(w, 75, -1);
        gtk_widget_show(w);
	gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
        /* Print command prompt. */
        w = (GtkWidget *)GUIPromptBar(
            NULL, "Print Command:",
            NULL, &entry
        );
	pw->output_printer_entry = (GtkWidget *)entry;
        GUISetWidgetTip((GtkWidget *)entry, PRINT_TTM_PRINT_COMMAND);
        gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
        gtk_widget_show(w);

        /* Hbox for print to file widgets. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent3 = w;
	/* Radio button. */
        pw->output_file_radio = w =
            gtk_radio_button_new_with_label(gslist, "File");
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "toggled",
            GTK_SIGNAL_FUNC(PrintWinToggledCB),
            (gpointer)pw
        );
        gtk_widget_set_usize(w, 75, -1);
        gtk_widget_show(w);
        /* Prompt with browse. */
        w = (GtkWidget *)GUIPromptBarWithBrowse(
            NULL, "Name:",
            NULL, &entry, &browse_btn,
            (gpointer)pw, PrintWinPrintFileBrowseCB
        );
        pw->output_file_entry = (GtkWidget *)entry;
	pw->output_file_browse_btn = (GtkWidget *)browse_btn;
        gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
        gtk_widget_show(w);


	/* Hbox to divide into two columns. */
	w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;


	/* Left side vbox. */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent3 = w;


	/* Visual. */
	/* Frame. */
	w = gtk_frame_new("Visual");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
        /* Vbox. */
        w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_show(w);
        parent4 = w;

        /* Vbox for brightness ticks and scale. */
        w = gtk_vbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Hbox for ticks. */
        w = gtk_hbox_new(FALSE, 0);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent6 = w;
        w = (GtkWidget *)GUICreateMenuItemIcon(
            (u_int8_t **)icon_minus_16x16_xpm
        );
        gtk_widget_set_usize(w, 16, 16);
        gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        w = (GtkWidget *)GUICreateMenuItemIcon(
            (u_int8_t **)icon_brightness_16x16_xpm
        );
        gtk_widget_set_usize(w, 16, 16);
        gtk_box_pack_start(GTK_BOX(parent6), w, TRUE, FALSE, 0);
        gtk_widget_show(w);
        w = (GtkWidget *)GUICreateMenuItemIcon(
            (u_int8_t **)icon_plus_16x16_xpm
        );
        gtk_widget_set_usize(w, 16, 16);
        gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        /* Brightness scale. */
        pw->brightness_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0,                /* Initial value. */
            -1.0,               /* Minimum. */
            1.0,                /* Maximum. */
            0.25,               /* Step inc. */
            0.50,               /* Page inc. */
            0.0                 /* Page size. */
        );
        /* Scale. */
        gtk_object_ref(GTK_OBJECT(adj));        /* Keep adj around. */
        pw->brightness_scale = w = gtk_hscale_new(adj);
        gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
        gtk_scale_set_value_pos(GTK_SCALE(w), GTK_POS_RIGHT);
        gtk_scale_set_digits(GTK_SCALE(w), 2);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(PrintWinAdjustmentValueChangedCB),
            (gpointer)pw
        );
        GUISetWidgetTip(w, PRINT_TTM_BRIGHTNESS);
        gtk_widget_show(w);

        /* Vbox for contrast ticks and scale. */
        w = gtk_vbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Hbox for ticks. */
        w = gtk_hbox_new(FALSE, 0);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent6 = w;
        w = (GtkWidget *)GUICreateMenuItemIcon(
            (u_int8_t **)icon_minus_16x16_xpm
        );
        gtk_widget_set_usize(w, 16, 16);
        gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        w = (GtkWidget *)GUICreateMenuItemIcon(
            (u_int8_t **)icon_contrast_16x16_xpm
        );
        gtk_widget_set_usize(w, 16, 16);
        gtk_box_pack_start(GTK_BOX(parent6), w, TRUE, FALSE, 0);
        gtk_widget_show(w);
        w = (GtkWidget *)GUICreateMenuItemIcon(
            (u_int8_t **)icon_plus_16x16_xpm
        );
        gtk_widget_set_usize(w, 16, 16);
        gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        /* Contrast scale. */
        pw->contrast_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0,                /* Initial value. */
            -1.0,               /* Minimum. */
            1.0,                /* Maximum. */
            0.25,               /* Step inc. */
            0.50,               /* Page inc. */
            0.0                 /* Page size. */
        );
        /* Scale. */
        gtk_object_ref(GTK_OBJECT(adj));        /* Keep adj around. */
        pw->contrast_scale = w = gtk_hscale_new(adj);
        gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
        gtk_scale_set_value_pos(GTK_SCALE(w), GTK_POS_RIGHT);
        gtk_scale_set_digits(GTK_SCALE(w), 2);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(PrintWinAdjustmentValueChangedCB),
            (gpointer)pw
        );
        GUISetWidgetTip(w, PRINT_TTM_CONTRAST);
        gtk_widget_show(w);


	/* Hbox for greyscale and color radios. */
	w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
	/* Greyscale radio. */
        gslist = NULL;
        pw->visual_greyscale_radio = w =
            gtk_radio_button_new_with_label(gslist, "GreyScale");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "toggled",
            GTK_SIGNAL_FUNC(PrintWinToggledCB),
            (gpointer)pw
        );
        gtk_widget_show(w);
        gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	/* Color radio. */
        pw->visual_color_radio = w =
            gtk_radio_button_new_with_label(gslist, "Color");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "toggled",
            GTK_SIGNAL_FUNC(PrintWinToggledCB),
            (gpointer)pw
        );
        gtk_widget_show(w);


        /* Orientation. */
        /* Frame. */
        w = gtk_frame_new("Orientation");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
        /* Hbox. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent4), w);
        gtk_widget_show(w);
        parent4 = w;
        /* Portrait radio. */
        gslist = NULL;
        pw->orientation_portrait_radio = w =
            gtk_radio_button_new_with_label(gslist, "Portrait");
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "toggled",
            GTK_SIGNAL_FUNC(PrintWinToggledCB),
            (gpointer)pw
        );
        gtk_widget_show(w);
        gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
        /* Landscape radio. */
        pw->orientation_landscape_radio = w =
            gtk_radio_button_new_with_label(gslist, "Landscape");
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "toggled",
            GTK_SIGNAL_FUNC(PrintWinToggledCB),
            (gpointer)pw
        );
        gtk_widget_show(w);

	/* Paper size. */
        w = gtk_frame_new("Paper Size");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
        gtk_widget_show(w);
        parent4 = w;
	/* Vbox. */
	w = gtk_vbox_new(FALSE, border_major);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent4), w);
        gtk_widget_show(w);
        parent4 = w;
	/* Scrolled window for clist. */
        w = gtk_scrolled_window_new(NULL, NULL);
        gtk_scrolled_window_set_policy(
            GTK_SCROLLED_WINDOW(w),
            GTK_POLICY_AUTOMATIC,
            GTK_POLICY_AUTOMATIC
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
        gtk_widget_show(w);
        scroll_parent = w;
	/* Clist. */
	heading[0] = "Type";
        heading[1] = "Size";
	pw->paper_size_clist = w = gtk_clist_new_with_titles(2, heading);
        gtk_widget_set_usize(w, 250, -1);
	gtk_clist_column_titles_hide(GTK_CLIST(w));
        gtk_clist_column_titles_passive(GTK_CLIST(w));
        gtk_clist_set_selection_mode(GTK_CLIST(w), GTK_SELECTION_SINGLE);
        gtk_clist_set_row_height(GTK_CLIST(w), PRINT_CLIST_ROW_SPACING);
        gtk_container_add(GTK_CONTAINER(scroll_parent), w);
        gtk_clist_set_column_justification(
            GTK_CLIST(w), 0, GTK_JUSTIFY_LEFT
        );
	gtk_clist_set_column_width(GTK_CLIST(w), 0, 90);
        gtk_clist_set_column_width(GTK_CLIST(w), 1, 10);
        gtk_clist_set_shadow_type(GTK_CLIST(w), GTK_SHADOW_IN);
        gtk_signal_connect(
            GTK_OBJECT(w), "select_row",
            GTK_SIGNAL_FUNC(PrintWinSelectRowCB),
            (gpointer)pw
        );
        gtk_widget_show(w);
	/* Add paper sizes to clist, itterate through paper_size list.
	 * The paper_size list contains a set of 4 pointers for each
	 * entry so we increment i in units of 4.
	 */
	for(i = 0; paper_size[i] != NULL; i += 4)
	{
	    gint row;
	    const gchar *cstrptr;
	    gchar **data = (gchar **)g_malloc0(
		4 * sizeof(gchar *)
	    );
	    if(data == NULL)
		break;

	    heading[0] = (gchar *)paper_size[i];
	    heading[1] = (gchar *)paper_size[i + 1];

	    /* Append a new row to the clist. */
	    row = gtk_clist_append(GTK_CLIST(w), heading);
	    if(row < 0)
	    {
		g_free(data);
		break;
	    }

	    /* Copy paper size entry to newly allocated row data, the
	     * first two pointers of the paper size entry are pointers
	     * to strings while the last two are gint values.
	     */
	    cstrptr = paper_size[i];			/* gchar *name */
	    if(cstrptr != NULL)
		data[0] = g_strdup(cstrptr);

	    cstrptr = paper_size[i + 1];		/* gchar *size */
	    if(cstrptr != NULL)
                data[1] = g_strdup(cstrptr);

	    data[2] = (gchar *)paper_size[i + 2];	/* gint width */
            data[3] = (gchar *)paper_size[i + 3];	/* gint height */

	    gtk_clist_set_row_data_full(
                GTK_CLIST(w), row,
		(gpointer)data, PrintWinPaperSizeRowDataDeleteCB
            );
	}

        /* Hbox for resolution. */
        w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* DPI X */
        w = (GtkWidget *)GUIPromptBar(
            NULL, "DPI X:",
            NULL, &entry
        );
        pw->dpi_x_entry = (GtkWidget *)entry;
        GUISetWidgetTip((GtkWidget *)entry, PRINT_TTM_DPI_X);
        gtk_widget_set_usize((GtkWidget *)entry, 50, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            (GtkObject *)entry, "changed",
            GTK_SIGNAL_FUNC(PrintWinDPIChangedCB),
            pw
        );
        gtk_widget_show(w);
        /* DPI Y */
        w = (GtkWidget *)GUIPromptBar(
            NULL, "DPI Y:",
            NULL, &entry
        );
        pw->dpi_y_entry = (GtkWidget *)entry;
        GUISetWidgetTip((GtkWidget *)entry, PRINT_TTM_DPI_Y);
        gtk_widget_set_usize((GtkWidget *)entry, 50, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            (GtkObject *)entry, "changed",
            GTK_SIGNAL_FUNC(PrintWinDPIChangedCB),
            pw
        );
        gtk_widget_show(w);




	/* Right side vbox. */
	w = gtk_vbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
        gtk_widget_show(w);
        parent3 = w;


        /* Output geometry. */
        w = gtk_frame_new("Output Geometry");
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
        /* Vbox. */
        w = gtk_vbox_new(FALSE, border_minor);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent4), w);
        gtk_widget_show(w);
        parent4 = w;
        /* Hbox. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
	/* Output X. */
        w = (GtkWidget *)GUIPromptBar(
            NULL, "X:",
            NULL, &entry
        );
        pw->output_x_entry = (GtkWidget *)entry;
        gtk_widget_set_usize((GtkWidget *)entry, 50, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            (GtkObject *)entry, "changed",
            GTK_SIGNAL_FUNC(PrintWinOutputEntryChangedCB),
            pw
        );
        gtk_widget_show(w);
        /* Output Y. */
        w = (GtkWidget *)GUIPromptBar(
            NULL, "Y:",
            NULL, &entry
        );
        pw->output_y_entry = (GtkWidget *)entry;
        gtk_widget_set_usize((GtkWidget *)entry, 50, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            (GtkObject *)entry, "changed",
            GTK_SIGNAL_FUNC(PrintWinOutputEntryChangedCB),
            pw
        );
        gtk_widget_show(w);
        /* Output width. */
        w = (GtkWidget *)GUIPromptBar(
            NULL, "Width:",
            NULL, &entry
        );
        pw->output_width_entry = (GtkWidget *)entry;
        gtk_widget_set_usize((GtkWidget *)entry, 50, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            (GtkObject *)entry, "changed",
            GTK_SIGNAL_FUNC(PrintWinOutputEntryChangedCB),
            pw
        );
        gtk_widget_show(w);
        /* Output height. */
        w = (GtkWidget *)GUIPromptBar(
            NULL, "Height:",
            NULL, &entry
        );
        pw->output_height_entry = (GtkWidget *)entry;
        gtk_widget_set_usize((GtkWidget *)entry, 50, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            (GtkObject *)entry, "changed",
            GTK_SIGNAL_FUNC(PrintWinOutputEntryChangedCB),
            pw
        );
        gtk_widget_show(w);


	/* Preview. */
	w = gtk_frame_new("Preview");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
        gtk_widget_show(w);
        parent4 = w;
        /* Vbox. */
        w = gtk_vbox_new(TRUE, border_minor);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent4), w);
        gtk_widget_show(w);
        parent4 = w;
	/* Hbox. */
        w = gtk_hbox_new(TRUE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
	/* Frame. */
	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_OUT);
	gtk_container_border_width(GTK_CONTAINER(w), border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
	/* Nother vbox. */
	w = gtk_vbox_new(TRUE, border_minor);
        gtk_container_add(GTK_CONTAINER(parent5), w);
        gtk_widget_show(w);
        parent5 = w;
	/* Drawing area. */
	pw->preview_da = w = gtk_drawing_area_new();
        /* Needs to accept focus for keyboard modifier events. */
        GTK_WIDGET_SET_FLAGS(w, GTK_CAN_FOCUS | GTK_CAN_DEFAULT);
        gtk_widget_add_events(
            w,
            GDK_EXPOSURE_MASK | GDK_ENTER_NOTIFY_MASK |
            GDK_LEAVE_NOTIFY_MASK | GDK_KEY_PRESS_MASK |
            GDK_KEY_RELEASE_MASK | GDK_BUTTON_PRESS_MASK |
            GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK |
            GDK_POINTER_MOTION_HINT_MASK | GDK_STRUCTURE_MASK
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "enter_notify_event",
            GTK_SIGNAL_FUNC(PrintWinPreviewEventCB),
            pw
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "leave_notify_event",
            GTK_SIGNAL_FUNC(PrintWinPreviewEventCB),
            pw
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "key_press_event",
            GTK_SIGNAL_FUNC(PrintWinPreviewEventCB),
            pw
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "key_release_event",
            GTK_SIGNAL_FUNC(PrintWinPreviewEventCB),
            pw
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "button_press_event",
            GTK_SIGNAL_FUNC(PrintWinPreviewEventCB),
            pw
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "button_release_event",
            GTK_SIGNAL_FUNC(PrintWinPreviewEventCB),
            pw
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "motion_notify_event",
            GTK_SIGNAL_FUNC(PrintWinPreviewEventCB),
            pw
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "expose_event",
            GTK_SIGNAL_FUNC(PrintWinPreviewExposeCB),
            pw
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "configure_event",
            GTK_SIGNAL_FUNC(PrintWinPreviewEventCB),
            pw
        );
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        GUISetWidgetTip(w, PRINT_TTM_PREVIEW);
        gtk_widget_show(w);

	/* Create preview graphic contexts. */
	style = gtk_widget_get_style(w);
	if(style != NULL)
	{
	    GdkGC *gc;

	    /* Use drawing area's base GC. */
	    gc = style->base_gc[GTK_STATE_NORMAL];
	    if(gc != NULL)
	    {
		gdk_gc_ref(gc);
		pw->gc_base = gc;
	    }
	}
        /* Create invert GC. */
        colormap = gtk_widget_get_colormap(w);
        if(colormap == NULL)
            colormap = gdk_colormap_get_system();
        if(colormap != NULL)
        {
            GdkColor *cf, *cb;
            GdkGCValues gcv;
            GdkGCValuesMask gcv_mask = (
                GDK_GC_FOREGROUND | GDK_GC_BACKGROUND | 
/* GDK_GC_FONT | */
                GDK_GC_FUNCTION | GDK_GC_FILL | GDK_GC_LINE_WIDTH |
                GDK_GC_LINE_STYLE | GDK_GC_CAP_STYLE | GDK_GC_JOIN_STYLE
            );
            cf = &gcv.foreground;
            cf->red = (guint16)(0.0 * (guint16)-1);
            cf->green = (guint16)(0.0 * (guint16)-1);
            cf->blue = (guint16)(0.0 * (guint16)-1);
            gdk_colormap_alloc_color(colormap, cf, TRUE, TRUE);
            cb = &gcv.background;
            cb->red = (guint16)(1.0 * (guint16)-1);
            cb->green = (guint16)(1.0 * (guint16)-1);
            cb->blue = (guint16)(1.0 * (guint16)-1);
            gdk_colormap_alloc_color(colormap, cb, TRUE, TRUE);
/*
            if(font != NULL)
                gdk_font_ref(font);
            gcv.font = font;
 */
            gcv.function = GDK_INVERT;
            gcv.fill = GDK_SOLID;
            gcv.line_width = 1;
            gcv.line_style = GDK_LINE_SOLID;
            gcv.cap_style = GDK_CAP_NOT_LAST;
            gcv.join_style = GDK_JOIN_MITER;
            pw->gc_invert = gdk_gc_new_with_values(
                window, &gcv, gcv_mask
            );
            gdk_colormap_free_colors(colormap, cf, 1);
            gdk_colormap_free_colors(colormap, cb, 1);
/*
            if(font != NULL)
                gdk_font_unref(font);
 */
	}


        /* Separator. */
        w = gtk_hseparator_new();
        gtk_box_pack_start(GTK_BOX(main_vbox), w, FALSE, FALSE, 0);
        gtk_widget_show(w);


        /* Buttons hbox. */
        w = gtk_hbox_new(TRUE, 0);
        gtk_box_pack_start(GTK_BOX(main_vbox), w, FALSE, FALSE, border_major);
        gtk_widget_show(w);
        parent2 = w;

        /* Print button. */
        pw->print_btn = w = GUIButtonPixmapLabelH(
            (u_int8_t **)icon_print_20x20_xpm, "Print", NULL
        );
        gtk_widget_set_usize(w, bw, bh);
        GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
        gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "clicked",
            GTK_SIGNAL_FUNC(PrintWinPrintCB),
            (gpointer)pw
        );
        gtk_accel_group_add(
            accelgrp, GDK_Return, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_accel_group_add(
            accelgrp, GDK_3270_Enter, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_accel_group_add(
            accelgrp, GDK_KP_Enter, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_accel_group_add(
            accelgrp, GDK_ISO_Enter, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
	gtk_widget_show(w);

        /* Cancel button. */
        pw->cancel_btn = w = GUIButtonPixmapLabelH(
            (u_int8_t **)icon_cancel_20x20_xpm, "Cancel", NULL
        );
        gtk_widget_set_usize(w, bw, bh);
        GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
        gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "clicked",
            GTK_SIGNAL_FUNC(PrintWinCancelCB),
            (gpointer)pw
        );
        gtk_accel_group_add(
            accelgrp, GDK_Escape, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_widget_show(w);




	/* Resize the preview drawing area. */
	PrintWinResizePreview(pw);

	/* Recreate the preview image. */
	PrintWinRecreatePreview(pw);

	return(pw);
}

/*
 *	Maps the print window.
 */
void PrintWinMap(print_win_struct *pw)
{
	GtkWidget *w;

	if(pw == NULL)
	    return;

	if(!pw->map_state)
	{
	    w = pw->print_btn;
	    if(w != NULL)
	    {
		gtk_widget_grab_focus(w);
		gtk_widget_grab_default(w);
	    }

	    w = pw->toplevel;
	    if(w != NULL)
		gtk_widget_show(w);

	    pw->map_state = TRUE;
	}
}

/*
 *	Unmaps the print window.
 */
void PrintWinUnmap(print_win_struct *pw)
{
        GtkWidget *w;

        if(pw == NULL)
            return;

        if(pw->map_state)
        {
            w = pw->toplevel;
            if(w != NULL)
                gtk_widget_hide(w);

            pw->map_state = FALSE;
        }
}

/*
 *	Destroys the given print window.
 */
void PrintWinDelete(print_win_struct *pw)
{
	GdkGC **gc;
	GdkCursor **cur;
	GtkAdjustment **adj;
	GtkWidget **w;


	if(pw == NULL)
	    return;

	if(pw->initialized)
	{
#define DO_UNREF_GC		\
{ \
 if((*gc) != NULL) \
 { \
  GdkGC *tgc = *gc; \
  (*gc) = NULL; \
  gdk_gc_unref(tgc); \
 } \
}
#define DO_DESTROY_CURSOR	\
{ \
 if((*cur) != NULL) \
 { \
  GdkCursor *tc = *cur; \
  (*cur) = NULL; \
  gdk_cursor_destroy(tc); \
 }\
}
#define DO_UNREF_ADJUSTMENT	\
{ \
 if((*adj) != NULL) \
 { \
  GtkObject *to = GTK_OBJECT(*adj); \
  (*adj) = NULL; \
  gtk_object_unref(to); \
 } \
}
#define DO_DESTROY_WIDGET       \
{ \
 if((*w) != NULL) \
 { \
  GtkWidget *tmp_w = *w; \
  (*w) = NULL; \
  gtk_widget_destroy(tmp_w); \
 } \
}
	    pw->iv = NULL;

	    /* Delete preview images. */
	    gdk_flush();
	    ImgViewImageDelete(pw->preview_img);
	    pw->preview_img = NULL;
	    ImgViewImageDelete(pw->orig_preview_img);
	    pw->orig_preview_img = NULL;

	    w = &pw->menu;
	    DO_DESTROY_WIDGET

	    w = &pw->paper_size_clist;
	    DO_DESTROY_WIDGET

	    w = &pw->toplevel;
	    pw->output_printer_radio = NULL;
	    pw->output_file_radio = NULL;
	    pw->output_printer_entry = NULL;
	    pw->output_file_entry = NULL;
	    pw->output_file_browse_btn = NULL;
	    pw->brightness_scale = NULL;
	    pw->contrast_scale = NULL;
	    pw->visual_greyscale_radio = NULL;
	    pw->visual_color_radio = NULL;
	    pw->orientation_portrait_radio = NULL;
	    pw->orientation_landscape_radio = NULL;
	    pw->dpi_x_entry = NULL;
	    pw->dpi_y_entry = NULL;
	    pw->output_x_entry = NULL;
	    pw->output_y_entry = NULL;
	    pw->output_width_entry = NULL;
	    pw->output_height_entry = NULL;
	    pw->preview_da = NULL;
	    pw->print_btn = NULL;
	    pw->cancel_btn = NULL;
	    DO_DESTROY_WIDGET

	    if(pw->accelgrp != NULL)
	    {
		gtk_accel_group_unref(pw->accelgrp);
		pw->accelgrp = NULL;
	    }

	    adj = &pw->brightness_adj;
	    DO_UNREF_ADJUSTMENT
            adj = &pw->contrast_adj;
            DO_UNREF_ADJUSTMENT

	    cur = &pw->busy_cur;
	    DO_DESTROY_CURSOR
	    cur = &pw->translate_cur;
	    DO_DESTROY_CURSOR
	    cur = &pw->size_cur;
	    DO_DESTROY_CURSOR

	    gc = &pw->gc_base;
	    DO_UNREF_GC
            gc = &pw->gc_invert;
            DO_UNREF_GC

#undef DO_DESTROY_WIDGET
#undef DO_UNREF_ADJUSTMENT
#undef DO_DESTROY_CURSOR
#undef DO_UNREF_GC
	}

	g_free(pw);
}


/*
 *	Creates a new image based on the given image and the print values.
 *
 *	The returned image must be deallocated by calling 
 *	ImgViewImageDelete().
 */
imgview_image_struct *PrintWinCreateImageFromValues(
	imgview_image_struct *src_img, print_values_struct *v
)
{
	guint vflags;
	gint orientation = PRINT_ORIENTATION_PORTRAIT;
	gint	output_width = 0,
		output_height = 0;
	imgview_image_struct *tar_img;


	if((src_img == NULL) || (v == NULL))
	    return(NULL);

        /* Get print values. */
        vflags = v->flags;
	if(vflags & PRINT_VALUE_ORIENTATION)
	    orientation = v->orientation;
	if(vflags & PRINT_VALUE_OUTPUT_WIDTH)
	    output_width = v->output_width;
	if(vflags & PRINT_VALUE_OUTPUT_HEIGHT)
	    output_height = v->output_height;

	/* Output size too small? */
	if((output_width < 1) || (output_height < 1))
	    return(NULL);

	/* Create target image depending on orientation. */
	if(orientation == PRINT_ORIENTATION_LANDSCAPE)
	{
	    /* Create tempory image that will be landscaled (with
	     * the width and height swapped).
	     */
	    imgview_image_struct *tmp_img = ImgViewImageNew(
	        src_img->height, src_img->width, src_img->depth
	    );
	    if(tmp_img == NULL)
		return(NULL);

	    /* Give up if bytes per pixel or geometry do not match.
	     * Remember that geometry is swapped since orientation is 
	     * landscape.
	     */
	    if((tmp_img->bpp != src_img->bpp) ||
	       (tmp_img->width != src_img->height) ||
	       (tmp_img->height != src_img->width)
	    )
            {
                ImgViewImageDelete(tmp_img);
                return(NULL);
            }

	    /* Copy source image data to target image. */
	    PrintWinLandscapeImage(
		src_img->bpp,
		(const guint8 *)src_img->mem,
		src_img->width, src_img->height, src_img->bpl,
		(guint8 *)tmp_img->mem,
		tmp_img->bpl
	    );

	    /* Create actual target image that is scaled. */
	    tar_img = ImgViewImageNew(
                output_width, output_height, src_img->depth
            );
            if(tar_img == NULL)
	    {
		ImgViewImageDelete(tmp_img);
                return(NULL);
	    }
	    /* Target image's bpp doesn't match tempory's bpp? */
            if(tar_img->bpp != tmp_img->bpp)
	    {
                ImgViewImageDelete(tmp_img);
		ImgViewImageDelete(tar_img);
                return(NULL);
            }

	    /* Copy and scale. */
	    GUIResizeBuffer(
		tar_img->bpp,
		(const gpointer)tmp_img->mem,
		tmp_img->width, tmp_img->height, tmp_img->bpl,
		(gpointer)tar_img->mem,
		tar_img->width, tar_img->height, tar_img->bpl
	    );

	    /* Delete tempory image, it is no longer needed. */
	    ImgViewImageDelete(tmp_img);
	}
	else
	{
	    /* Create target image that is scailed in size. */
            tar_img = ImgViewImageNew(
                output_width, output_height, src_img->depth
            );
            if(tar_img == NULL)
                return(NULL);

            /* Target image's bpp doesn't match source's bpp? */
            if(tar_img->bpp != src_img->bpp)
	    {
		ImgViewImageDelete(tar_img);
                return(NULL);
	    }

            /* Copy and scale. */
            GUIResizeBuffer(
                src_img->bpp,
                (const gpointer)src_img->mem,
                src_img->width, src_img->height, src_img->bpl,
                (gpointer)tar_img->mem,
                tar_img->width, tar_img->height, tar_img->bpl
            );
	}

	/* Update new target image's visual. */
	PrintWinUpdatePreviewVisualV(tar_img, v);

	return(tar_img);
}



/*
 *	Creates a new print window and blocks output untill a response
 *	is recieved. The given values in the print_values_struct will be
 *	used to initialize the values and the print_values_struct will be
 *	modified if TRUE is returned. Otherwise print_values_struct *v
 *	should be considered garbage after this call.
 */
gbool PrintWinGetResponse(
	imgview_struct *iv,
        GtkWidget *ref_toplevel,	/* Reference toplevel. */
	print_values_struct *v		/* May not be NULL. */
)
{
	GtkWidget *w;
	guint vflags;
	print_win_struct *pw;
	static gchar lfile_name[PATH_MAX + NAME_MAX];
#define PRINT_COMMAND_MAX	((PATH_MAX + NAME_MAX) * 2) + 1024
	static gchar lprint_command[PRINT_COMMAND_MAX];


	/* Must have print values structure. */
	if(v == NULL)
	    return(FALSE);

	/* Get print values flags. */
	vflags = v->flags;

	/* Create new print window. */
	pw = PrintWinNew(iv);
	if(pw == NULL)
	    return(FALSE);

	/* Reset globals. */
	got_response = FALSE;
	main_block_level = 0;


	/* Begin setting values. */

        /* Output image geometry. */
        if(vflags & PRINT_VALUE_OUTPUT_X)
            pw->output_x = v->output_x;
        if(vflags & PRINT_VALUE_OUTPUT_Y)
            pw->output_y = v->output_y;
        if(vflags & PRINT_VALUE_OUTPUT_WIDTH)
            pw->output_width = v->output_width;
        if(vflags & PRINT_VALUE_OUTPUT_HEIGHT)
            pw->output_height = v->output_height;
	/* Update output entry widgets with the new output geometry
	 * values.
	 */
	PrintWinSyncOutputWidgets(pw);

	/* Print command. */
	if(vflags & PRINT_VALUE_PRINT_COMMAND)
	{
            w = pw->output_printer_entry;
            if((w != NULL) ? GTK_IS_ENTRY(w) : 0)
                gtk_entry_set_text(
		    GTK_ENTRY(w),
		    (v->print_command != NULL) ? v->print_command : ""
		);
	}
	/* File name. */
        if(vflags & PRINT_VALUE_FILE_NAME)
        {
            w = pw->output_file_entry;
            if((w != NULL) ? GTK_IS_ENTRY(w) : 0)
                gtk_entry_set_text(
                    GTK_ENTRY(w),
                    (v->file_name != NULL) ? v->file_name : ""
                );
        }
	/* Print output type. */
	if(vflags & PRINT_VALUE_OUTPUT_TYPE)
	{
	    if(v->output_type == PRINT_OUTPUT_TO_FILE)
	    {
	        w = pw->output_printer_radio;
	        if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : 0)
		    GTK_TOGGLE_BUTTON(w)->active = FALSE;
                w = pw->output_file_radio;
                if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : 0)
                    GTK_TOGGLE_BUTTON(w)->active = TRUE;

                w = pw->output_file_radio;
                if(w != NULL)
                    gtk_signal_emit_by_name(GTK_OBJECT(w), "toggled");
	    }
	    else
	    {
                w = pw->output_printer_radio;
                if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : 0)
                    GTK_TOGGLE_BUTTON(w)->active = TRUE;
                w = pw->output_file_radio;
                if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : 0)
                    GTK_TOGGLE_BUTTON(w)->active = FALSE;

                w = pw->output_printer_radio;
                if(w != NULL)
                    gtk_signal_emit_by_name(GTK_OBJECT(w), "toggled");
	    }
        }
	/* Brightness. */
	if(vflags & PRINT_VALUE_BRIGHTNESS)
	{
	    GtkAdjustment *adj = pw->brightness_adj;
	    if(adj != NULL)
	    {
		adj->value = v->brightness;
		if(adj->value < adj->lower)
		    adj->value = adj->lower;
		if(adj->value > adj->upper)
		    adj->value = adj->upper;
		gtk_signal_emit_by_name(GTK_OBJECT(adj), "changed");
	    }
	}
        /* Contrast. */
        if(vflags & PRINT_VALUE_CONTRAST)
        {
            GtkAdjustment *adj = pw->contrast_adj;
            if(adj != NULL)
            {
                adj->value = v->contrast;
                if(adj->value < adj->lower)
                    adj->value = adj->lower;
                if(adj->value > adj->upper)
                    adj->value = adj->upper;
                gtk_signal_emit_by_name(GTK_OBJECT(adj), "changed");
            }
        }
	/* Visual. */
	if(vflags & PRINT_VALUE_VISUAL)
	{
	    w = pw->visual_greyscale_radio;
            if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : 0)
                GTK_TOGGLE_BUTTON(w)->active = FALSE;
            w = pw->visual_color_radio;
            if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : 0)
                GTK_TOGGLE_BUTTON(w)->active = FALSE;

	    switch(v->visual)
	    {
	      case PRINT_VISUAL_GREYSCALE:
		w = pw->visual_greyscale_radio;
		break;
	      case PRINT_VISUAL_COLOR:
                w = pw->visual_color_radio;
                break;
	    }
	    if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : 0)
	    {
                GTK_TOGGLE_BUTTON(w)->active = TRUE;
/*		gtk_signal_emit_by_name(GTK_OBJECT(w), "toggled"); */
	    }
	}
        /* Orientation. */
        if(vflags & PRINT_VALUE_ORIENTATION)
	{
            w = pw->orientation_portrait_radio;
            if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : 0)
                GTK_TOGGLE_BUTTON(w)->active = FALSE;
            w = pw->orientation_landscape_radio;
            if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : 0)
                GTK_TOGGLE_BUTTON(w)->active = FALSE;

            switch(v->orientation)
            {
              case PRINT_ORIENTATION_PORTRAIT:
                w = pw->orientation_portrait_radio;
                break;
              case PRINT_ORIENTATION_LANDSCAPE:
                w = pw->orientation_landscape_radio;
                break;
            }
            if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : 0)
            {
                GTK_TOGGLE_BUTTON(w)->active = TRUE;
/*		gtk_signal_emit_by_name(GTK_OBJECT(w), "toggled"); */
            }
        }
	/* DPI X. */
	if(vflags & PRINT_VALUE_DPI_X)
	{
	    gchar num_str[80];

	    w = pw->dpi_x_entry;
	    if(w != NULL)
	    {
		sprintf(num_str, "%i", v->dpi_x);
		gtk_entry_set_text(GTK_ENTRY(w), num_str);
	    }
	}
        /* DPI Y. */
        if(vflags & PRINT_VALUE_DPI_Y)
        {
            gchar num_str[80];

            w = pw->dpi_y_entry;
            if(w != NULL)
            {
                sprintf(num_str, "%i", v->dpi_y);
                gtk_entry_set_text(GTK_ENTRY(w), num_str);
            }
        }
	/* Paper size. */
	if((vflags & PRINT_VALUE_PAPER_WIDTH) &&
           (vflags & PRINT_VALUE_PAPER_HEIGHT)
	)
	{
	    gint row, width, height;
	    const gchar **data;
	    GtkCList *clist = (GtkCList *)pw->paper_size_clist;
	    if(clist != NULL)
	    {
		/* Itterate through each row on the clist and get
		 * its row data. Check row data to see if its size matches
		 * with the given paper size and if it does then select
		 * the row in question and break out of the loop.
		 */
		for(row = 0; row < clist->rows; row++)
		{
		    data = (const gchar **)gtk_clist_get_row_data(clist, row);
		    if(data == NULL)
			continue;

		    /* Get last two pointers of this entry which are
		     * gint's.
		     */
		    width = (gint)data[2];
		    height = (gint)data[3];

		    if((v->paper_width == width) &&
		       (v->paper_height == height)
		    )
		    {
			gtk_clist_select_row(clist, row, 0);
			break;
		    }
		}
		/* No row selected? */
		if(row >= clist->rows)
		{
		    /* select first row if possible. */
		    if(clist->rows > 0)
			gtk_clist_select_row(clist, 0, 0);
		}
	    }
	}
	/* Note that selecting a paper size above with 
	 * gtk_clist_select_row() would have called the "select_row"
	 * callback and update the preview widget and preview image.
	 */




	/* Set reference toplevel? */
	if((ref_toplevel != NULL) && (pw->toplevel != NULL))
	{
	    if(GTK_IS_WINDOW(ref_toplevel))
	    {
/* Can't really set it as modal otherwise file browser that is used gets
   locked out.
 */
	        gtk_window_set_modal(
		    GTK_WINDOW(pw->toplevel), TRUE
	        );
	        gtk_window_set_transient_for(
		    GTK_WINDOW(pw->toplevel), GTK_WINDOW(ref_toplevel)
	        );
	    }
	}


	/* Map print window. */
	PrintWinMap(pw);

	/* Increment GTK+ main block level and enter GTK+ main loop. */
	main_block_level++;
	gtk_main();

	/* Break out of any remaining GTK+ main levels. */
	while(main_block_level > 0)
	{
	    main_block_level--;
	    gtk_main_quit();
	}


	/* Got response? */
	if(got_response)
	{
	    /* Collect values and update returns. */

	    /* Print output type. */
	    if(vflags & PRINT_VALUE_OUTPUT_TYPE)
	    {
		w = pw->output_file_radio;
		if((w != NULL) ? GTK_TOGGLE_BUTTON(w)->active : 0)
		    v->output_type = PRINT_OUTPUT_TO_FILE;
		else
		    v->output_type = PRINT_OUTPUT_TO_PRINTER;
	    }
	    /* Print command. */
	    if(vflags & PRINT_VALUE_PRINT_COMMAND)
	    {
		w = pw->output_printer_entry;
		if(w != NULL)
		{
		    const gchar *cstrptr = gtk_entry_get_text(GTK_ENTRY(w));
		    if(cstrptr != NULL)
		    {
			strncpy(lprint_command, cstrptr, PRINT_COMMAND_MAX);
			lprint_command[PRINT_COMMAND_MAX - 1] = '\0';
			v->print_command = lprint_command;
		    }
		}
	    }
            /* Print to file name. */
	    if(vflags & PRINT_VALUE_FILE_NAME)
	    {
                w = pw->output_file_entry;
                if(w != NULL)
                {
                    const gchar *cstrptr = gtk_entry_get_text(GTK_ENTRY(w));
                    if(cstrptr != NULL)
                    {
                        strncpy(lfile_name, cstrptr, PATH_MAX + NAME_MAX);
                        lfile_name[PATH_MAX + NAME_MAX - 1] = '\0';
                        v->file_name = lfile_name;
                    }
                }
            }
	    /* Brightness. */
	    if(vflags & PRINT_VALUE_BRIGHTNESS)
	    {
		GtkAdjustment *adj = pw->brightness_adj;
		if(adj != NULL)
		{
		    v->brightness = adj->value;
		}
	    }
            /* Contrast. */
            if(vflags & PRINT_VALUE_CONTRAST)
            {
                GtkAdjustment *adj = pw->contrast_adj;
                if(adj != NULL)
                {
                    v->contrast = adj->value;
                }
            }
	    /* Visual. */
	    if(vflags & PRINT_VALUE_VISUAL)
	    {
		w = pw->visual_greyscale_radio;
		if((w != NULL) ? GTK_TOGGLE_BUTTON(w)->active : 0)
		    v->visual = PRINT_VISUAL_GREYSCALE;
		else
		    v->visual = PRINT_VISUAL_COLOR;
	    }
            /* Orientation. */
            if(vflags & PRINT_VALUE_ORIENTATION)
            {
                w = pw->orientation_landscape_radio;
                if((w != NULL) ? GTK_TOGGLE_BUTTON(w)->active : 0)
                    v->orientation = PRINT_ORIENTATION_LANDSCAPE;
                else
                    v->orientation = PRINT_ORIENTATION_PORTRAIT;
            }
            /* DPI X. */
	    if(vflags & PRINT_VALUE_DPI_X)
	    {
		w = pw->dpi_x_entry;
		if((w != NULL) ? GTK_IS_ENTRY(w) : 0)
                {
		    const gchar *cstrptr = gtk_entry_get_text(GTK_ENTRY(w));
		    if(cstrptr != NULL)
			v->dpi_x = atoi(cstrptr);
		}
            }
            /* DPI Y. */
            if(vflags & PRINT_VALUE_DPI_Y)
            {
                w = pw->dpi_y_entry;
                if((w != NULL) ? GTK_IS_ENTRY(w) : 0)
                {
                    const gchar *cstrptr = gtk_entry_get_text(GTK_ENTRY(w));
                    if(cstrptr != NULL)
                        v->dpi_y = atoi(cstrptr);
                }
            }
	    /* Paper size. */
	    if((vflags & PRINT_VALUE_PAPER_WIDTH) &&
               (vflags & PRINT_VALUE_PAPER_HEIGHT)
	    )
	    {
		/* Get paper size (without DPI applied). */
		PrintWinPaperSize(pw, &v->paper_width, &v->paper_height, FALSE);
	    }

	    /* Output image geometry. */
	    if(vflags & PRINT_VALUE_OUTPUT_X)
		v->output_x = pw->output_x;
            if(vflags & PRINT_VALUE_OUTPUT_Y)
                v->output_y = pw->output_y;
            if(vflags & PRINT_VALUE_OUTPUT_WIDTH)
                v->output_width = pw->output_width;
            if(vflags & PRINT_VALUE_OUTPUT_HEIGHT)
                v->output_height = pw->output_height;
	}


	/* Unset reference toplevel. */
	if(pw->toplevel != NULL)
	{
	    gtk_window_set_modal(
		GTK_WINDOW(pw->toplevel), FALSE
	    );
	    gtk_window_set_transient_for(
		GTK_WINDOW(pw->toplevel), NULL
	    );
	}

	/* Destroy the print window. */
	PrintWinDelete(pw);
	pw = NULL;

#undef PRINT_COMMAND_MAX

	return(got_response);
}
